#!/usr/bin/env python3

import copy
import io
import os
import random
import ruamel.yaml
import shutil
import unittest
import tempfile
import textwrap
from collections import OrderedDict
from pathlib import Path
from unittest import mock

import fdroidserver
from fdroidserver import metadata
from fdroidserver.exception import MetaDataException
from fdroidserver.common import DEFAULT_LOCALE
from .testcommon import TmpCwd, mkdtemp


basedir = Path(__file__).parent


def _get_mock_mf(s):
    mf = io.StringIO(s)
    mf.name = 'mock_filename.yaml'
    return mf


class MetadataTest(unittest.TestCase):
    '''fdroidserver/metadata.py'''

    def setUp(self):
        os.chdir(basedir)
        self._td = mkdtemp()
        self.testdir = self._td.name
        fdroidserver.metadata.warnings_action = 'error'

    def tearDown(self):
        # auto-generated dirs by functions, not tests, so they are not always cleaned up
        self._td.cleanup()
        try:
            os.rmdir("srclibs")
        except OSError:
            pass
        try:
            os.rmdir("tmp")
        except OSError:
            pass

    def test_fieldtypes_key_exist(self):
        for k in fdroidserver.metadata.fieldtypes:
            self.assertIn(k, fdroidserver.metadata.yaml_app_fields)

    def test_build_flagtypes_key_exist(self):
        for k in fdroidserver.metadata.flagtypes:
            self.assertIn(k, fdroidserver.metadata.build_flags)

    def test_FieldValidator_BitcoinAddress(self):
        validator = None
        for vali in fdroidserver.metadata.valuetypes:
            if vali.name == 'Bitcoin address':
                validator = vali
                break
        self.assertIsNotNone(validator, "could not find 'Bitcoin address' validator")

        # some valid addresses (P2PKH, P2SH, Bech32)
        self.assertIsNone(
            validator.check('1BrrrrErsrWetrTrnrrrrm4GFg7xJaNVN2', 'fake.app.id')
        )
        self.assertIsNone(
            validator.check('3JrrrrWrEZr3rNrrvrecrnyirrnqRhWNLy', 'fake.app.id')
        )
        self.assertIsNone(
            validator.check('bc1qar0srrr7xrkvr5lr43lrdnwrre5rgtrzrf5rrq', 'fake.app.id')
        )

        # some invalid addresses
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            validator.check,
            '21BvMrSYsrWrtrrlL5A10mlGFr7rrarrN2',
            'fake.app.id',
        )
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            validator.check,
            '5Hrgr3ur5rGLrfKrrrrrrHSrqJrroGrrzrQrrrrrrLNrsrDrrrA',
            'fake.app.id',
        )
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            validator.check,
            '92rr46rUrgTrrromrVrirW6r1rrrdrerrdbJrrrhrCsYrrrrrrc',
            'fake.app.id',
        )
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            validator.check,
            'K1BvMrSYsrWrtrrrn5Au4m4GFr7rrarrN2',
            'fake.app.id',
        )
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            validator.check,
            'L1BvMrSYsrWrtrrrn5Au4m4GFr7rrarrN2',
            'fake.app.id',
        )
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            validator.check,
            'tb1qw5r8drrejxrrg4y5rrrrrraryrrrrwrkxrjrsx',
            'fake.app.id',
        )

    def test_FieldValidator_LitecoinAddress(self):
        validator = None
        for vali in fdroidserver.metadata.valuetypes:
            if vali.name == 'Litecoin address':
                validator = vali
                break
        self.assertIsNotNone(validator, "could not find 'Litecoin address' validator")

        # some valid addresses (L, M, 3, segwit)
        self.assertIsNone(
            validator.check('LgeGrrrrJAxyXprrPrrBrrX5Qrrrrrrrrd', 'fake.app.id')
        )
        self.assertIsNone(
            validator.check('MrrrrrrrJAxyXpanPtrrRAX5QHxvUJo8id', 'fake.app.id')
        )
        self.assertIsNone(validator.check('3rereVr9rAryrranrrrrrAXrrHx', 'fake.app.id'))
        self.assertIsNone(
            validator.check(
                'ltc1q7euacwhn6ef99vcfa57mute92q572aqsc4c2j5', 'fake.app.id'
            )
        )

        # some invalid addresses (various special use/testnet addresses, invalid chars)
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            validator.check,
            '21BvMrSYsrWrtrrrn5Au4l4GFr7rrarrN2',
            'fake.app.id',
        )
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            validator.check,
            '5Hrgr3ur5rGLrfKrrrrrr1SrqJrroGrrzrQrrrrrrLNrsrDrrrA',
            'fake.app.id',
        )
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            validator.check,
            '92rr46rUrgTrrromrVrirW6r1rrrdrerrdbJrrrhrCsYrrrrrrc',
            'fake.app.id',
        )
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            validator.check,
            'K1BvMrSYsrWrtrrrn5Au4m4GFr7rrarrN2',
            'fake.app.id',
        )
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            validator.check,
            'L0000rSYsrWrtrrrn5Au4m4GFr7rrarrN2',
            'fake.app.id',
        )
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            validator.check,
            'tb1qw5r8drrejxrrg4y5rrrrrraryrrrrwrkxrjrsx',
            'fake.app.id',
        )

    def test_valid_funding_yml_regex(self):
        """Check the regex can find all the cases"""
        with (basedir / 'funding-usernames.yaml').open() as fp:
            yaml = ruamel.yaml.YAML(typ='safe')
            data = yaml.load(fp)

        for k, entries in data.items():
            for entry in entries:
                m = fdroidserver.metadata.VALID_USERNAME_REGEX.match(entry)
                if k == 'custom':
                    pass
                elif k == 'bad':
                    self.assertIsNone(
                        m, 'this is an invalid %s username: {%s}' % (k, entry)
                    )
                else:
                    self.assertIsNotNone(
                        m, 'this is a valid %s username: {%s}' % (k, entry)
                    )

    @mock.patch('git.Repo', mock.Mock())
    @mock.patch('logging.error')
    def test_read_metadata(self, logging_error):
        """Read specified metadata files included in tests/, compare to stored output"""

        self.maxDiff = None

        config = dict()
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.common.config = config
        fdroidserver.metadata.warnings_action = None

        yaml = ruamel.yaml.YAML(typ='safe')
        apps = fdroidserver.metadata.read_metadata()
        for appid in (
            'app.with.special.build.params',
            'org.smssecure.smssecure',
            'org.adaway',
            'org.videolan.vlc',
            'com.politedroid',
        ):
            savepath = Path('metadata/dump') / (appid + '.yaml')
            frommeta = dict(apps[appid])
            self.assertTrue(appid in apps)
            with savepath.open('r') as f:
                from_yaml = yaml.load(f)
            self.assertEqual(frommeta, from_yaml)
            # comment above assert and uncomment below to update test
            # files when new metadata fields are added
            # with savepath.open('w') as fp:
            #     yaml.default_flow_style = False
            #     yaml.register_class(metadata.Build)
            #     yaml.dump(frommeta, fp)

        # errors are printed when .yml overrides localized
        logging_error.assert_called()
        self.assertEqual(3, len(logging_error.call_args_list))

    @mock.patch('git.Repo', mock.Mock())
    def test_metadata_overrides_dot_fdroid_yml(self):
        """Fields in metadata files should override anything in .fdroid.yml."""
        app = metadata.parse_metadata('metadata/info.guardianproject.urzip.yml')
        self.assertEqual(app['Summary'], '一个实用工具，获取已安装在您的设备上的应用的有关信息')

    def test_dot_fdroid_yml_works_without_git(self):
        """Parsing should work if .fdroid.yml is present and it is not a git repo."""
        os.chdir(self.testdir)
        yml = Path('metadata/test.yml')
        yml.parent.mkdir()
        with yml.open('w') as fp:
            fp.write('Repo: https://example.com/not/git/or/anything')
        fdroid_yml = Path('build/test/.fdroid.yml')
        fdroid_yml.parent.mkdir(parents=True)
        with fdroid_yml.open('w') as fp:
            fp.write('OpenCollective: test')
        metadata.parse_metadata(yml)  # should not throw an exception

    @mock.patch('git.Repo', mock.Mock())
    @mock.patch('logging.error')
    def test_rewrite_yaml_fakeotaupdate(self, logging_error):
        with tempfile.TemporaryDirectory() as testdir:
            testdir = Path(testdir)
            fdroidserver.common.config = {'accepted_formats': ['yml']}
            fdroidserver.metadata.warnings_action = None

            # rewrite metadata
            allapps = fdroidserver.metadata.read_metadata()
            for appid, app in allapps.items():
                if appid == 'fake.ota.update':
                    fdroidserver.metadata.write_metadata(
                        testdir / (appid + '.yml'), app
                    )

            # assert rewrite result
            self.maxDiff = None
            file_name = 'fake.ota.update.yml'
            self.assertEqual(
                (testdir / file_name).read_text(encoding='utf-8'),
                (Path('metadata-rewrite-yml') / file_name).read_text(encoding='utf-8'),
            )

        # errors are printed when .yml overrides localized
        logging_error.assert_called()
        self.assertEqual(3, len(logging_error.call_args_list))

    @mock.patch('git.Repo', mock.Mock())
    def test_rewrite_yaml_fdroidclient(self):
        with tempfile.TemporaryDirectory() as testdir:
            testdir = Path(testdir)
            fdroidserver.common.config = {'accepted_formats': ['yml']}

            # rewrite metadata
            allapps = fdroidserver.metadata.read_metadata()
            for appid, app in allapps.items():
                if appid == 'org.fdroid.fdroid':
                    fdroidserver.metadata.write_metadata(
                        testdir / (appid + '.yml'), app
                    )

            # assert rewrite result
            self.maxDiff = None
            file_name = 'org.fdroid.fdroid.yml'
            self.assertEqual(
                (testdir / file_name).read_text(encoding='utf-8'),
                (Path('metadata-rewrite-yml') / file_name).read_text(encoding='utf-8'),
            )

    @mock.patch('git.Repo', mock.Mock())
    def test_rewrite_yaml_special_build_params(self):
        """Test rewriting a plain YAML metadata file without localized files."""
        os.chdir(self.testdir)
        os.mkdir('metadata')
        appid = 'app.with.special.build.params'
        file_name = Path('metadata/%s.yml' % appid)
        shutil.copy(basedir / file_name, file_name)

        # rewrite metadata
        allapps = fdroidserver.metadata.read_metadata({appid: -1})
        for appid, app in allapps.items():
            metadata.write_metadata(file_name, app)

        # assert rewrite result
        self.maxDiff = None
        self.assertEqual(
            file_name.read_text(),
            (basedir / 'metadata-rewrite-yml' / file_name.name).read_text(),
        )

    def test_normalize_type_string(self):
        """TYPE_STRING currently has some quirky behavior."""
        self.assertEqual('123456', metadata._normalize_type_string(123456))
        self.assertEqual('1.0', metadata._normalize_type_string(1.0))
        self.assertEqual('0', metadata._normalize_type_string(0))
        self.assertEqual('0.0', metadata._normalize_type_string(0.0))
        self.assertEqual('0.1', metadata._normalize_type_string(0.1))
        self.assertEqual('[]', metadata._normalize_type_string(list()))
        self.assertEqual('{}', metadata._normalize_type_string(dict()))
        self.assertEqual('false', metadata._normalize_type_string(False))
        self.assertEqual('true', metadata._normalize_type_string(True))

    def test_normalize_type_string_sha256(self):
        """SHA-256 values are TYPE_STRING, which YAML can parse as decimal ints."""
        yaml = ruamel.yaml.YAML(typ='safe')
        for v in range(1, 1000):
            s = '%064d' % (v * (10**51))
            self.assertEqual(s, metadata._normalize_type_string(yaml.load(s)))

    def test_normalize_type_stringmap_none(self):
        self.assertEqual(dict(), metadata._normalize_type_stringmap('key', None))

    def test_normalize_type_stringmap_empty_list(self):
        self.assertEqual(dict(), metadata._normalize_type_stringmap('AntiFeatures', []))

    def test_normalize_type_stringmap_simple_list_format(self):
        self.assertEqual(
            {'Ads': {}, 'Tracking': {}},
            metadata._normalize_type_stringmap('AntiFeatures', ['Ads', 'Tracking']),
        )

    def test_normalize_type_int(self):
        """TYPE_INT should be an int whenever possible."""
        self.assertEqual(0, metadata._normalize_type_int('key', 0))
        self.assertEqual(1, metadata._normalize_type_int('key', 1))
        self.assertEqual(-5, metadata._normalize_type_int('key', -5))
        self.assertEqual(0, metadata._normalize_type_int('key', '0'))
        self.assertEqual(1, metadata._normalize_type_int('key', '1'))
        self.assertEqual(-5, metadata._normalize_type_int('key', '-5'))
        self.assertEqual(
            12345678901234567890,
            metadata._normalize_type_int('key', 12345678901234567890),
        )

    def test_normalize_type_int_fails(self):
        with self.assertRaises(MetaDataException):
            metadata._normalize_type_int('key', '1a')
        with self.assertRaises(MetaDataException):
            metadata._normalize_type_int('key', 1.1)
        with self.assertRaises(MetaDataException):
            metadata._normalize_type_int('key', True)

    def test_normalize_type_list(self):
        """TYPE_LIST is always a list of strings, no matter what YAML thinks."""
        k = 'placeholder'
        yaml = ruamel.yaml.YAML(typ='safe')
        self.assertEqual(['1.0'], metadata._normalize_type_list(k, 1.0))
        self.assertEqual(['1234567890'], metadata._normalize_type_list(k, 1234567890))
        self.assertEqual(['false'], metadata._normalize_type_list(k, False))
        self.assertEqual(['true'], metadata._normalize_type_list(k, True))
        self.assertEqual(['foo'], metadata._normalize_type_list(k, 'foo'))
        self.assertEqual([], metadata._normalize_type_list(k, list()))
        self.assertEqual([], metadata._normalize_type_list(k, tuple()))
        self.assertEqual([], metadata._normalize_type_list(k, set()))
        self.assertEqual(['0', '1', '2'], metadata._normalize_type_list(k, {0, 1, 2}))
        self.assertEqual(
            ['a', 'b', 'c', '0', '0.0'],
            metadata._normalize_type_list(k, yaml.load('[a, b, c, 0, 0.0]')),
        )
        self.assertEqual(
            ['1', '1.0', 's', 'true', '{}'],
            metadata._normalize_type_list(k, yaml.load('[1, 1.0, s, true, {}]')),
        )
        self.assertEqual(
            ['1', '1.0', 's', 'true', '{}'],
            metadata._normalize_type_list(k, (1, 1.0, 's', True, dict())),
        )

    def test_normalize_type_list_fails(self):
        with self.assertRaises(MetaDataException):
            metadata._normalize_type_list('placeholder', dict())

    def test_post_parse_yaml_metadata(self):
        yamldata = dict()
        metadata.post_parse_yaml_metadata(yamldata)

        yamldata[
            'AllowedAPKSigningKeys'
        ] = 'c03dac71394d6c26766f1b04d3e31cfcac5d03b55d8aa40cc9b9fa6b74354c66'
        metadata.post_parse_yaml_metadata(yamldata)

    def test_post_parse_yaml_metadata_ArchivePolicy_int(self):
        for i in range(20):
            yamldata = {'ArchivePolicy': i}
            metadata.post_parse_yaml_metadata(yamldata)
            self.assertEqual(i, yamldata['ArchivePolicy'])

    def test_post_parse_yaml_metadata_ArchivePolicy_string(self):
        for i in range(20):
            yamldata = {'ArchivePolicy': '%d' % i}
            metadata.post_parse_yaml_metadata(yamldata)
            self.assertEqual(i, yamldata['ArchivePolicy'])

    def test_post_parse_yaml_metadata_ArchivePolicy_versions(self):
        """Test that the old format still works."""
        for i in range(20):
            yamldata = {'ArchivePolicy': '%d versions' % i}
            metadata.post_parse_yaml_metadata(yamldata)
            self.assertEqual(i, yamldata['ArchivePolicy'])

    def test_post_parse_yaml_metadata_fails(self):
        yamldata = {'AllowedAPKSigningKeys': {'bad': 'dict-placement'}}
        with self.assertRaises(MetaDataException):
            metadata.post_parse_yaml_metadata(yamldata)

    def test_post_parse_yaml_metadata_0padding_sha256(self):
        """SHA-256 values are strings, but YAML 1.2 will read some as decimal ints."""
        v = '0027293472934293872934729834729834729834729834792837487293847926'
        yaml = ruamel.yaml.YAML(typ='safe')
        yamldata = yaml.load('AllowedAPKSigningKeys: ' + v)
        metadata.post_parse_yaml_metadata(yamldata)
        self.assertEqual(yamldata['AllowedAPKSigningKeys'], [v])

    def test_post_parse_yaml_metadata_builds(self):
        yamldata = OrderedDict()
        builds = []
        yamldata['Builds'] = builds
        build = OrderedDict()
        builds.append(build)

        build['versionCode'] = 1.1
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            fdroidserver.metadata.post_parse_yaml_metadata,
            yamldata,
        )

        build['versionCode'] = '1a'
        self.assertRaises(
            fdroidserver.exception.MetaDataException,
            fdroidserver.metadata.post_parse_yaml_metadata,
            yamldata,
        )

        build['versionCode'] = 1
        build['versionName'] = 1
        fdroidserver.metadata.post_parse_yaml_metadata(yamldata)
        self.assertNotEqual(1, yamldata['Builds'][0]['versionName'])
        self.assertEqual('1', yamldata['Builds'][0]['versionName'])
        self.assertEqual(1, yamldata['Builds'][0]['versionCode'])

        build['versionName'] = 1.0
        fdroidserver.metadata.post_parse_yaml_metadata(yamldata)
        self.assertNotEqual(1.0, yamldata['Builds'][0]['versionName'])
        self.assertEqual('1.0', yamldata['Builds'][0]['versionName'])

        build['commit'] = 1.0
        fdroidserver.metadata.post_parse_yaml_metadata(yamldata)
        self.assertNotEqual(1.0, yamldata['Builds'][0]['commit'])
        self.assertEqual('1.0', yamldata['Builds'][0]['commit'])

        teststr = '98234fab134b'
        build['commit'] = teststr
        fdroidserver.metadata.post_parse_yaml_metadata(yamldata)
        self.assertEqual(teststr, yamldata['Builds'][0]['commit'])

        testcommitid = 1234567890
        build['commit'] = testcommitid
        fdroidserver.metadata.post_parse_yaml_metadata(yamldata)
        self.assertNotEqual(testcommitid, yamldata['Builds'][0]['commit'])
        self.assertEqual('1234567890', yamldata['Builds'][0]['commit'])

    def test_read_metadata_sort_by_time(self):
        with tempfile.TemporaryDirectory() as testdir, TmpCwd(testdir):
            testdir = Path(testdir)
            metadatadir = testdir / 'metadata'
            metadatadir.mkdir()

            randomlist = []
            randomapps = list((basedir / 'metadata').glob('*.yml'))
            random.shuffle(randomapps)
            i = 1
            for f in randomapps:
                shutil.copy(f, metadatadir)
                new = metadatadir / f.name
                stat = new.stat()
                os.utime(new, (stat.st_ctime, stat.st_mtime + i))
                # prepend new item so newest is always first
                randomlist = [f.stem] + randomlist
                i += 1
            allapps = fdroidserver.metadata.read_metadata(sort_by_time=True)
            allappids = []
            for appid, app in allapps.items():
                allappids.append(appid)
            self.assertEqual(randomlist, allappids)

    def test_parse_yaml_metadata_0size_file(self):
        self.assertEqual(dict(), metadata.parse_yaml_metadata(_get_mock_mf('')))

    def test_parse_yaml_metadata_empty_dict_file(self):
        self.assertEqual(dict(), metadata.parse_yaml_metadata(_get_mock_mf('{}')))

    def test_parse_yaml_metadata_empty_string_file(self):
        self.assertEqual(dict(), metadata.parse_yaml_metadata(_get_mock_mf('""')))

    def test_parse_yaml_metadata_fail_on_root_list(self):
        with self.assertRaises(MetaDataException):
            metadata.parse_yaml_metadata(_get_mock_mf('-'))
        with self.assertRaises(MetaDataException):
            metadata.parse_yaml_metadata(_get_mock_mf('[]'))
        with self.assertRaises(MetaDataException):
            metadata.parse_yaml_metadata(_get_mock_mf('- AutoName: fake'))

    def test_parse_yaml_metadata_type_list_str(self):
        v = 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
        mf = _get_mock_mf('AllowedAPKSigningKeys: "%s"' % v)
        self.assertEqual(
            v,
            metadata.parse_yaml_metadata(mf)['AllowedAPKSigningKeys'][0],
        )

    def test_parse_yaml_metadata_type_list_build_str(self):
        mf = _get_mock_mf('Builds: [{versionCode: 1, rm: s}]')
        self.assertEqual(
            metadata.parse_yaml_metadata(mf),
            {'Builds': [{'rm': ['s'], 'versionCode': 1}]},
        )

    def test_parse_yaml_metadata_app_type_list_fails(self):
        mf = _get_mock_mf('AllowedAPKSigningKeys: {t: f}')
        with self.assertRaises(MetaDataException):
            metadata.parse_yaml_metadata(mf)

    def test_parse_yaml_metadata_build_type_list_fails(self):
        mf = _get_mock_mf('Builds: [{versionCode: 1, rm: {bad: dict-placement}}]')
        with self.assertRaises(MetaDataException):
            metadata.parse_yaml_metadata(mf)

    def test_parse_yaml_metadata_unknown_app_field(self):
        mf = io.StringIO(
            textwrap.dedent(
                """\
                AutoName: F-Droid
                RepoType: git
                Builds: []
                bad: value"""
            )
        )
        mf.name = 'mock_filename.yaml'
        with self.assertRaises(MetaDataException):
            fdroidserver.metadata.parse_yaml_metadata(mf)

    def test_parse_yaml_metadata_unknown_build_flag(self):
        mf = io.StringIO(
            textwrap.dedent(
                """\
                AutoName: F-Droid
                RepoType: git
                Builds:
                - bad: value"""
            )
        )
        mf.name = 'mock_filename.yaml'
        with self.assertRaises(MetaDataException):
            fdroidserver.metadata.parse_yaml_metadata(mf)

    @mock.patch('logging.warning')
    @mock.patch('logging.error')
    def test_parse_yaml_metadata_continue_on_warning(self, _error, _warning):
        """When errors are disabled, parsing should provide something that can work.

        When errors are disabled, then it should try to give data that
        lets something happen.  A zero-length file is valid for
        operation, it just declares a Application ID as "known" and
        nothing else.  This example gives a list as the base in the
        .yml file, which is unparsable, so it gives a warning message
        and carries on with a blank dict.

        """
        fdroidserver.metadata.warnings_action = None
        mf = _get_mock_mf('[AntiFeatures: Tracking]')
        self.assertEqual(fdroidserver.metadata.parse_yaml_metadata(mf), dict())
        _warning.assert_called_once()
        _error.assert_called_once()

    def test_parse_localized_antifeatures(self):
        """Unit test based on reading files included in the test repo."""
        app = dict()
        app['id'] = 'app.with.special.build.params'
        metadata.parse_localized_antifeatures(app)
        self.maxDiff = None
        self.assertEqual(
            app,
            {
                'AntiFeatures': {
                    'Ads': {'en-US': 'please no'},
                    'NoSourceSince': {'en-US': 'no activity\n'},
                },
                'Builds': [
                    {
                        'versionCode': 50,
                        'antifeatures': {
                            'Ads': {
                                'en-US': 'includes ad lib\n',
                                'zh-CN': '包括广告图书馆\n',
                            },
                            'Tracking': {'en-US': 'standard suspects\n'},
                        },
                    },
                    {
                        'versionCode': 49,
                        'antifeatures': {
                            'Tracking': {'zh-CN': 'Text from zh-CN/49_Tracking.txt'},
                        },
                    },
                ],
                'id': app['id'],
            },
        )

    def test_parse_localized_antifeatures_passthrough(self):
        """Test app values are cleanly passed through if no localized files."""
        before = {
            'id': 'placeholder',
            'AntiFeatures': {'NonFreeDep': {}},
            'Builds': [{'versionCode': 999, 'antifeatures': {'zero': {}, 'one': {}}}],
        }
        after = copy.deepcopy(before)
        with tempfile.TemporaryDirectory() as testdir:
            os.chdir(testdir)
            os.mkdir('metadata')
            os.mkdir(os.path.join('metadata', after['id']))
            metadata.parse_localized_antifeatures(after)
        self.assertEqual(before, after)

    def test_parse_metadata_antifeatures_NoSourceSince(self):
        """Test that NoSourceSince gets added as an Anti-Feature."""
        os.chdir(self.testdir)
        yml = Path('metadata/test.yml')
        yml.parent.mkdir()
        with yml.open('w') as fp:
            fp.write('AntiFeatures: Ads\nNoSourceSince: gone\n')
        app = metadata.parse_metadata(yml)
        self.assertEqual(
            app['AntiFeatures'], {'Ads': {}, 'NoSourceSince': {DEFAULT_LOCALE: 'gone'}}
        )

    @mock.patch('logging.error')
    def test_yml_overrides_localized_antifeatures(self, logging_error):
        """Definitions in .yml files should override the localized versions."""
        app = metadata.parse_metadata('metadata/app.with.special.build.params.yml')

        self.assertEqual(app['AntiFeatures'], {'UpstreamNonFree': {}})

        self.assertEqual(49, app['Builds'][-3]['versionCode'])
        self.assertEqual(
            app['Builds'][-3]['antifeatures'],
            {'Tracking': {DEFAULT_LOCALE: 'Uses the Facebook SDK.'}},
        )

        self.assertEqual(50, app['Builds'][-2]['versionCode'])
        self.assertEqual(
            app['Builds'][-2]['antifeatures'],
            {
                'Ads': {
                    'en-US': 'includes ad lib\n',
                    'zh-CN': '包括广告图书馆\n',
                },
                'Tracking': {'en-US': 'standard suspects\n'},
            },
        )
        # errors are printed when .yml overrides localized
        logging_error.assert_called()
        self.assertEqual(3, len(logging_error.call_args_list))

    def test_parse_yaml_srclib_corrupt_file(self):
        with tempfile.TemporaryDirectory() as testdir:
            testdir = Path(testdir)
            srclibfile = testdir / 'srclib/mock.yml'
            srclibfile.parent.mkdir()
            with srclibfile.open('w') as fp:
                fp.write(
                    textwrap.dedent(
                        """
                        - RepoType: git
                        - Repo: https://github.com/realm/realm-js.git
                        """
                    )
                )
            with self.assertRaises(MetaDataException):
                fdroidserver.metadata.parse_yaml_srclib(srclibfile)

    def test_write_yaml_with_placeholder_values(self):
        mf = io.StringIO()

        app = fdroidserver.metadata.App()
        app.Categories = ['None']
        app.SourceCode = "https://gitlab.com/fdroid/fdroidclient.git"
        app.IssueTracker = "https://gitlab.com/fdroid/fdroidclient/issues"
        app.RepoType = 'git'
        app.Repo = 'https://gitlab.com/fdroid/fdroidclient.git'
        app.AutoUpdateMode = 'None'
        app.UpdateCheckMode = 'Tags'
        build = fdroidserver.metadata.Build()
        build.versionName = 'Unknown'  # taken from fdroidserver/import.py
        build.versionCode = 0  # taken from fdroidserver/import.py
        build.disable = 'Generated by import.py ...'
        build.commit = 'Unknown'
        build.gradle = ['yes']
        app['Builds'] = [build]

        fdroidserver.metadata.write_yaml(mf, app)

        mf.seek(0)
        self.assertEqual(
            mf.read(),
            textwrap.dedent(
                """\
                Categories:
                  - None
                License: Unknown
                SourceCode: https://gitlab.com/fdroid/fdroidclient.git
                IssueTracker: https://gitlab.com/fdroid/fdroidclient/issues

                RepoType: git
                Repo: https://gitlab.com/fdroid/fdroidclient.git

                Builds:
                  - versionName: Unknown
                    versionCode: 0
                    disable: Generated by import.py ...
                    commit: Unknown
                    gradle:
                      - yes

                AutoUpdateMode: None
                UpdateCheckMode: Tags
                """
            ),
        )

    def test_parse_yaml_metadata_prebuild_list(self):
        mf = io.StringIO(
            textwrap.dedent(
                """\
                AutoName: F-Droid
                RepoType: git
                Builds:
                  - versionCode: 1
                    versionName: v0.1.0
                    sudo:
                      - apt-get update
                      - apt-get install -y whatever
                      - sed -i -e 's/<that attr="bad"/<that attr="good"/' ~/.whatever/config.xml
                    init:
                      - bash generate_some_file.sh
                      - sed -i -e 'g/what/ever/' /some/file
                    prebuild:
                      - npm something
                      - echo 'important setting' >> /a/file
                    build:
                      - ./gradlew someSpecialTask
                      - sed -i 'd/that wrong config/' gradle.properties
                      - ./gradlew compile
                """
            )
        )
        mf.name = 'mock_filename.yaml'
        mf.seek(0)
        result = fdroidserver.metadata.parse_yaml_metadata(mf)
        self.maxDiff = None
        self.assertDictEqual(
            result,
            {
                'AutoName': 'F-Droid',
                'RepoType': 'git',
                'Builds': [
                    {
                        'versionCode': 1,
                        'versionName': 'v0.1.0',
                        'sudo': [
                            "apt-get update",
                            "apt-get install -y whatever",
                            "sed -i -e 's/<that attr=\"bad\"/<that attr=\"good\"/' ~/.whatever/config.xml",
                        ],
                        'init': [
                            "bash generate_some_file.sh",
                            "sed -i -e 'g/what/ever/' /some/file",
                        ],
                        'prebuild': [
                            "npm something",
                            "echo 'important setting' >> /a/file",
                        ],
                        'build': [
                            "./gradlew someSpecialTask",
                            "sed -i 'd/that wrong config/' gradle.properties",
                            "./gradlew compile",
                        ],
                    }
                ],
            },
        )

    def test_parse_yaml_metadata_prebuild_strings(self):
        mf = io.StringIO(
            textwrap.dedent(
                """\
                AutoName: F-Droid
                RepoType: git
                Builds:
                  - versionCode: 1
                    versionName: v0.1.0
                    sudo: |-
                      apt-get update && apt-get install -y whatever && sed -i -e 's/<that attr="bad"/<that attr="good"/' ~/.whatever/config.xml
                    init: bash generate_some_file.sh && sed -i -e 'g/what/ever/' /some/file
                    prebuild: npm something && echo 'important setting' >> /a/file
                    build: |-
                      ./gradlew someSpecialTask && sed -i 'd/that wrong config/' gradle.properties && ./gradlew compile
                """
            )
        )
        mf.name = 'mock_filename.yaml'
        mf.seek(0)
        result = fdroidserver.metadata.parse_yaml_metadata(mf)
        self.maxDiff = None
        self.assertDictEqual(
            result,
            {
                'AutoName': 'F-Droid',
                'RepoType': 'git',
                'Builds': [
                    {
                        'versionCode': 1,
                        'versionName': 'v0.1.0',
                        'sudo': [
                            "apt-get update && "
                            "apt-get install -y whatever && "
                            "sed -i -e 's/<that attr=\"bad\"/<that attr=\"good\"/' ~/.whatever/config.xml"
                        ],
                        'init': [
                            "bash generate_some_file.sh && "
                            "sed -i -e 'g/what/ever/' /some/file"
                        ],
                        'prebuild': [
                            "npm something && echo 'important setting' >> /a/file"
                        ],
                        'build': [
                            "./gradlew someSpecialTask && "
                            "sed -i 'd/that wrong config/' gradle.properties && "
                            "./gradlew compile"
                        ],
                    }
                ],
            },
        )

    def test_parse_yaml_metadata_prebuild_string(self):
        mf = io.StringIO(
            textwrap.dedent(
                """\
                AutoName: F-Droid
                RepoType: git
                Builds:
                  - versionCode: 1
                    versionName: v0.1.0
                    prebuild: |-
                      a && b && sed -i 's,a,b,'
                """
            )
        )
        mf.name = 'mock_filename.yaml'
        mf.seek(0)
        result = fdroidserver.metadata.parse_yaml_metadata(mf)
        self.assertDictEqual(
            result,
            {
                'AutoName': 'F-Droid',
                'RepoType': 'git',
                'Builds': [
                    {
                        'versionCode': 1,
                        'versionName': 'v0.1.0',
                        'prebuild': ["a && b && sed -i 's,a,b,'"],
                    }
                ],
            },
        )

    def test_parse_yaml_provides_should_be_ignored(self):
        mf = io.StringIO(
            textwrap.dedent(
                """\
                Provides: this.is.deprecated
                AutoName: F-Droid
                RepoType: git
                Builds:
                  - versionCode: 1
                    versionName: v0.1.0
                    prebuild: |-
                      a && b && sed -i 's,a,b,'
                """
            )
        )
        mf.name = 'mock_filename.yaml'
        mf.seek(0)
        result = fdroidserver.metadata.parse_yaml_metadata(mf)
        self.assertNotIn('Provides', result)
        self.assertNotIn('provides', result)

    def test_parse_yaml_app_antifeatures_dict(self):
        nonfreenet = 'free it!'
        tracking = 'so many'
        mf = io.StringIO(
            textwrap.dedent(
                f"""
                AntiFeatures:
                  Tracking: {tracking}
                  NonFreeNet: {nonfreenet}
                """
            )
        )
        self.assertEqual(
            metadata.parse_yaml_metadata(mf),
            {
                'AntiFeatures': {
                    'NonFreeNet': {DEFAULT_LOCALE: nonfreenet},
                    'Tracking': {DEFAULT_LOCALE: tracking},
                }
            },
        )

    def test_parse_yaml_metadata_build_antifeatures_old_style(self):
        mf = _get_mock_mf(
            textwrap.dedent(
                """
                AntiFeatures:
                  - Ads
                Builds:
                  - versionCode: 123
                    antifeatures:
                      - KnownVuln
                      - UpstreamNonFree
                      - NonFreeAssets
                """
            )
        )
        self.assertEqual(
            metadata.parse_yaml_metadata(mf),
            {
                'AntiFeatures': {'Ads': {}},
                'Builds': [
                    {
                        'antifeatures': {
                            'KnownVuln': {},
                            'NonFreeAssets': {},
                            'UpstreamNonFree': {},
                        },
                        'versionCode': 123,
                    }
                ],
            },
        )

    def test_parse_yaml_metadata_antifeatures_sort(self):
        """All data should end up sorted, to minimize diffs in the index files."""
        self.assertEqual(
            metadata.parse_yaml_metadata(
                _get_mock_mf(
                    textwrap.dedent(
                        """
                Builds:
                  - versionCode: 123
                    antifeatures:
                      KnownVuln:
                        es: 2nd
                        az: zero
                        en-US: first
                      UpstreamNonFree:
                      NonFreeAssets:
                AntiFeatures:
                  NonFreeDep:
                  Ads:
                    sw: 2nd
                    zh-CN: 3rd
                    de: 1st
                """
                    )
                )
            ),
            {
                'AntiFeatures': {
                    'Ads': {'de': '1st', 'sw': '2nd', 'zh-CN': '3rd'},
                    'NonFreeDep': {},
                },
                'Builds': [
                    {
                        'antifeatures': {
                            'KnownVuln': {'az': 'zero', 'en-US': 'first', 'es': '2nd'},
                            'NonFreeAssets': {},
                            'UpstreamNonFree': {},
                        },
                        'versionCode': 123,
                    }
                ],
            },
        )

    def test_parse_yaml_app_antifeatures_str(self):
        self.assertEqual(
            metadata.parse_yaml_metadata(io.StringIO('AntiFeatures: Tracking')),
            {'AntiFeatures': {'Tracking': {}}},
        )

    def test_parse_yaml_app_antifeatures_bool(self):
        self.assertEqual(
            metadata.parse_yaml_metadata(io.StringIO('AntiFeatures: true')),
            {'AntiFeatures': {'true': {}}},
        )

    def test_parse_yaml_app_antifeatures_float_nan(self):
        self.assertEqual(
            metadata.parse_yaml_metadata(io.StringIO('AntiFeatures: .nan')),
            {'AntiFeatures': {'.nan': {}}},
        )

    def test_parse_yaml_app_antifeatures_float_inf(self):
        self.assertEqual(
            metadata.parse_yaml_metadata(io.StringIO('AntiFeatures: .inf')),
            {'AntiFeatures': {'.inf': {}}},
        )

    def test_parse_yaml_app_antifeatures_float_negative_inf(self):
        self.assertEqual(
            metadata.parse_yaml_metadata(io.StringIO('AntiFeatures: -.inf')),
            {'AntiFeatures': {'-.inf': {}}},
        )

    def test_parse_yaml_app_antifeatures_int(self):
        self.assertEqual(
            metadata.parse_yaml_metadata(io.StringIO('AntiFeatures: 1')),
            {'AntiFeatures': {'1': {}}},
        )

    def test_parse_yaml_app_antifeatures_float(self):
        self.assertEqual(
            metadata.parse_yaml_metadata(io.StringIO('AntiFeatures: 1.0')),
            {'AntiFeatures': {'1.0': {}}},
        )

    def test_parse_yaml_app_antifeatures_list_float(self):
        self.assertEqual(
            metadata.parse_yaml_metadata(io.StringIO('AntiFeatures:\n  - 1.0\n')),
            {'AntiFeatures': {'1.0': {}}},
        )

    def test_parse_yaml_app_antifeatures_dict_float(self):
        mf = io.StringIO('AntiFeatures:\n  0.0: too early\n')
        self.assertEqual(
            metadata.parse_yaml_metadata(mf),
            {'AntiFeatures': {'0.0': {'en-US': 'too early'}}},
        )

    def test_parse_yaml_app_antifeatures_dict_float_fail_value(self):
        mf = io.StringIO('AntiFeatures:\n  NoSourceSince: 1.0\n')
        self.assertEqual(
            metadata.parse_yaml_metadata(mf),
            {'AntiFeatures': {'NoSourceSince': {'en-US': '1.0'}}},
        )

    def test_parse_yaml_metadata_type_stringmap_old_list(self):
        mf = _get_mock_mf(
            textwrap.dedent(
                """
                    AntiFeatures:
                      - Ads
                      - Tracking
                """
            )
        )
        self.assertEqual(
            {'AntiFeatures': {'Ads': {}, 'Tracking': {}}},
            metadata.parse_yaml_metadata(mf),
        )

    def test_parse_yaml_app_antifeatures_dict_no_value(self):
        mf = io.StringIO(
            textwrap.dedent(
                """\
                AntiFeatures:
                  Tracking:
                  NonFreeNet:
                """
            )
        )
        self.assertEqual(
            metadata.parse_yaml_metadata(mf),
            {'AntiFeatures': {'NonFreeNet': {}, 'Tracking': {}}},
        )

    def test_parse_yaml_metadata_type_stringmap_transitional(self):
        """Support a transitional format, where users just append a text"""
        ads = 'Has ad lib in it.'
        tracking = 'opt-out reports with ACRA'
        mf = _get_mock_mf(
            textwrap.dedent(
                f"""
                    AntiFeatures:
                      - Ads: {ads}
                      - Tracking: {tracking}
                """
            )
        )
        self.assertEqual(
            metadata.parse_yaml_metadata(mf),
            {
                'AntiFeatures': {
                    'Ads': {DEFAULT_LOCALE: ads},
                    'Tracking': {DEFAULT_LOCALE: tracking},
                }
            },
        )

    def test_parse_yaml_app_antifeatures_dict_mixed_values(self):
        ads = 'true'
        tracking = 'many'
        nonfreenet = '1'
        mf = io.StringIO(
            textwrap.dedent(
                f"""
                AntiFeatures:
                  Ads: {ads}
                  Tracking: {tracking}
                  NonFreeNet: {nonfreenet}
                """
            )
        )
        self.assertEqual(
            metadata.parse_yaml_metadata(mf),
            {
                'AntiFeatures': {
                    'Ads': {DEFAULT_LOCALE: ads},
                    'NonFreeNet': {DEFAULT_LOCALE: nonfreenet},
                    'Tracking': {DEFAULT_LOCALE: tracking},
                }
            },
        )

    def test_parse_yaml_app_antifeatures_stringmap_full(self):
        ads = 'watching'
        tracking = 'many'
        nonfreenet = 'pipes'
        nonfreenet_zh = '非免费网络'
        self.maxDiff = None
        mf = io.StringIO(
            textwrap.dedent(
                f"""
                AntiFeatures:
                  Ads:
                    {DEFAULT_LOCALE}: {ads}
                  Tracking:
                    {DEFAULT_LOCALE}: {tracking}
                  NonFreeNet:
                    {DEFAULT_LOCALE}: {nonfreenet}
                    zh-CN: {nonfreenet_zh}
                """
            )
        )
        self.assertEqual(
            metadata.parse_yaml_metadata(mf),
            {
                'AntiFeatures': {
                    'Ads': {DEFAULT_LOCALE: ads},
                    'NonFreeNet': {DEFAULT_LOCALE: nonfreenet, 'zh-CN': nonfreenet_zh},
                    'Tracking': {DEFAULT_LOCALE: tracking},
                }
            },
        )

    def test_parse_yaml_build_type_int_fail(self):
        mf = io.StringIO('Builds: [{versionCode: 1a}]')
        with self.assertRaises(MetaDataException):
            fdroidserver.metadata.parse_yaml_metadata(mf)

    def test_parse_yaml_int_strict_typing_fails(self):
        """Things that cannot be preserved when parsing as YAML."""
        mf = io.StringIO('Builds: [{versionCode: 1, rm: 0xf}]')
        self.assertEqual(
            {'Builds': [{'rm': ['15'], 'versionCode': 1}]},  # 15 != 0xf
            fdroidserver.metadata.parse_yaml_metadata(mf),
        )
        mf = io.StringIO('Builds: [{versionCode: 1, rm: 0x010}]')
        self.assertEqual(
            {'Builds': [{'rm': ['16'], 'versionCode': 1}]},  # 16 != 0x010
            fdroidserver.metadata.parse_yaml_metadata(mf),
        )
        mf = io.StringIO('Builds: [{versionCode: 1, rm: 0o015}]')
        self.assertEqual(
            {'Builds': [{'rm': ['13'], 'versionCode': 1}]},  # 13 != 0o015
            fdroidserver.metadata.parse_yaml_metadata(mf),
        )
        mf = io.StringIO('Builds: [{versionCode: 1, rm: 10_000}]')
        self.assertEqual(
            {'Builds': [{'rm': ['10000'], 'versionCode': 1}]},  # 10000 != 10_000
            fdroidserver.metadata.parse_yaml_metadata(mf),
        )

    def test_write_yaml_1_line_scripts_as_string(self):
        mf = io.StringIO()
        app = fdroidserver.metadata.App()
        app.Categories = ['None']
        app['Builds'] = []
        build = fdroidserver.metadata.Build()
        build.versionCode = 102030
        build.versionName = 'v1.2.3'
        build.sudo = ["chmod +rwx /opt"]
        build.init = ["sed -i -e 'g/what/ever/' /some/file"]
        build.prebuild = ["sed -i 'd/that wrong config/' gradle.properties"]
        build.build = ["./gradlew compile"]
        app['Builds'].append(build)
        fdroidserver.metadata.write_yaml(mf, app)
        mf.seek(0)
        self.assertEqual(
            mf.read(),
            textwrap.dedent(
                """\
                Categories:
                  - None
                License: Unknown

                Builds:
                  - versionName: v1.2.3
                    versionCode: 102030
                    sudo: chmod +rwx /opt
                    init: sed -i -e 'g/what/ever/' /some/file
                    prebuild: sed -i 'd/that wrong config/' gradle.properties
                    build: ./gradlew compile

                AutoUpdateMode: None
                UpdateCheckMode: None
                """
            ),
        )

    def test_write_yaml_1_line_scripts_as_list(self):
        mf = io.StringIO()
        app = fdroidserver.metadata.App()
        app.Categories = ['None']
        app['Builds'] = []
        build = fdroidserver.metadata.Build()
        build.versionCode = 102030
        build.versionName = 'v1.2.3'
        build.sudo = ["chmod +rwx /opt"]
        build.init = ["sed -i -e 'g/what/ever/' /some/file"]
        build.prebuild = ["sed -i 'd/that wrong config/' gradle.properties"]
        build.build = ["./gradlew compile"]
        app['Builds'].append(build)
        fdroidserver.metadata.write_yaml(mf, app)
        mf.seek(0)
        self.assertEqual(
            mf.read(),
            textwrap.dedent(
                """\
                Categories:
                  - None
                License: Unknown

                Builds:
                  - versionName: v1.2.3
                    versionCode: 102030
                    sudo: chmod +rwx /opt
                    init: sed -i -e 'g/what/ever/' /some/file
                    prebuild: sed -i 'd/that wrong config/' gradle.properties
                    build: ./gradlew compile

                AutoUpdateMode: None
                UpdateCheckMode: None
                """
            ),
        )

    def test_write_yaml_multiline_scripts_from_list(self):
        mf = io.StringIO()
        app = fdroidserver.metadata.App()
        app.Categories = ['None']
        app['Builds'] = []
        build = fdroidserver.metadata.Build()
        build.versionCode = 102030
        build.versionName = 'v1.2.3'
        build.sudo = [
            "apt-get update",
            "apt-get install -y whatever",
            "sed -i -e 's/<that attr=\"bad\"/<that attr=\"good\"/' ~/.whatever/config.xml",
        ]
        build.init = [
            "bash generate_some_file.sh",
            "sed -i -e 'g/what/ever/' /some/file",
        ]
        build.prebuild = ["npm something", "echo 'important setting' >> /a/file"]
        build.build = [
            "./gradlew someSpecialTask",
            "sed -i 'd/that wrong config/' gradle.properties",
            "./gradlew compile",
        ]
        app['Builds'].append(build)
        fdroidserver.metadata.write_yaml(mf, app)
        mf.seek(0)
        self.assertEqual(
            mf.read(),
            textwrap.dedent(
                """\
            Categories:
              - None
            License: Unknown

            Builds:
              - versionName: v1.2.3
                versionCode: 102030
                sudo:
                  - apt-get update
                  - apt-get install -y whatever
                  - sed -i -e 's/<that attr="bad"/<that attr="good"/' ~/.whatever/config.xml
                init:
                  - bash generate_some_file.sh
                  - sed -i -e 'g/what/ever/' /some/file
                prebuild:
                  - npm something
                  - echo 'important setting' >> /a/file
                build:
                  - ./gradlew someSpecialTask
                  - sed -i 'd/that wrong config/' gradle.properties
                  - ./gradlew compile

            AutoUpdateMode: None
            UpdateCheckMode: None
            """
            ),
        )

    def test_write_yaml_multiline_scripts_from_string(self):
        mf = io.StringIO()
        app = fdroidserver.metadata.App()
        app.Categories = ['None']
        app['Builds'] = []
        build = fdroidserver.metadata.Build()
        build.versionCode = 102030
        build.versionName = 'v1.2.3'
        build.sudo = [
            "apt-get update",
            "apt-get install -y whatever",
            "sed -i -e 's/<that attr=\"bad\"/<that attr=\"good\"/' ~/.whatever/config.xml",
        ]
        build.init = [
            "bash generate_some_file.sh",
            "sed -i -e 'g/what/ever/' /some/file",
        ]
        build.prebuild = ["npm something", "echo 'important setting' >> /a/file"]
        build.build = [
            "./gradlew someSpecialTask",
            "sed -i 'd/that wrong config/' gradle.properties",
            "./gradlew compile",
        ]
        app['Builds'].append(build)
        fdroidserver.metadata.write_yaml(mf, app)
        mf.seek(0)
        self.assertEqual(
            mf.read(),
            textwrap.dedent(
                """\
            Categories:
              - None
            License: Unknown

            Builds:
              - versionName: v1.2.3
                versionCode: 102030
                sudo:
                  - apt-get update
                  - apt-get install -y whatever
                  - sed -i -e 's/<that attr="bad"/<that attr="good"/' ~/.whatever/config.xml
                init:
                  - bash generate_some_file.sh
                  - sed -i -e 'g/what/ever/' /some/file
                prebuild:
                  - npm something
                  - echo 'important setting' >> /a/file
                build:
                  - ./gradlew someSpecialTask
                  - sed -i 'd/that wrong config/' gradle.properties
                  - ./gradlew compile

            AutoUpdateMode: None
            UpdateCheckMode: None
            """
            ),
        )

    def test_write_yaml_build_antifeatures(self):
        mf = io.StringIO()
        app = metadata.App(
            {
                'License': 'Apache-2.0',
                'Builds': [
                    metadata.Build(
                        {
                            'versionCode': 102030,
                            'versionName': 'v1.2.3',
                            'gradle': ['yes'],
                            'antifeatures': {
                                'a': {},
                                'b': {'de': 'Probe', 'en-US': 'test'},
                            },
                        }
                    ),
                ],
                'id': 'placeholder',
            }
        )
        metadata.write_yaml(mf, app)
        mf.seek(0)
        self.assertEqual(
            mf.read(),
            textwrap.dedent(
                """\
                License: Apache-2.0

                Builds:
                  - versionName: v1.2.3
                    versionCode: 102030
                    gradle:
                      - yes
                    antifeatures:
                      a: {}
                      b:
                        de: Probe
                        en-US: test
                """
            ),
        )

    def test_write_yaml_build_antifeatures_old_style(self):
        mf = io.StringIO()
        app = metadata.App(
            {
                'License': 'Apache-2.0',
                'Builds': [
                    metadata.Build(
                        {
                            'versionCode': 102030,
                            'versionName': 'v1.2.3',
                            'gradle': ['yes'],
                            'antifeatures': {'b': {}, 'a': {}},
                        }
                    ),
                ],
                'id': 'placeholder',
            }
        )
        metadata.write_yaml(mf, app)
        mf.seek(0)
        self.assertEqual(
            mf.read(),
            textwrap.dedent(
                """\
                License: Apache-2.0

                Builds:
                  - versionName: v1.2.3
                    versionCode: 102030
                    gradle:
                      - yes
                    antifeatures:
                      - a
                      - b
                """
            ),
        )

    def test_write_yaml_make_sure_provides_does_not_get_written(self):
        mf = io.StringIO()
        app = fdroidserver.metadata.App()
        app.Categories = ['None']
        app.Provides = 'this.is.deprecated'
        app['Builds'] = []
        build = fdroidserver.metadata.Build()
        build.versionCode = 102030
        build.versionName = 'v1.2.3'
        build.gradle = ['yes']
        app['Builds'].append(build)
        fdroidserver.metadata.write_yaml(mf, app)
        mf.seek(0)
        self.assertEqual(
            mf.read(),
            textwrap.dedent(
                """\
                Categories:
                  - None
                License: Unknown

                Builds:
                  - versionName: v1.2.3
                    versionCode: 102030
                    gradle:
                      - yes

                AutoUpdateMode: None
                UpdateCheckMode: None
                """
            ),
        )

    def test_parse_yaml_srclib_unknown_key(self):
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            with Path('test.yml').open('w', encoding='utf-8') as f:
                f.write(
                    textwrap.dedent(
                        '''\
                        RepoType: git
                        Repo: https://example.com/test.git
                        Evil: I should not be here.
                        '''
                    )
                )
            with self.assertRaisesRegex(
                MetaDataException,
                "Invalid srclib metadata: unknown key 'Evil' in 'test.yml'",
            ):
                fdroidserver.metadata.parse_yaml_srclib(Path('test.yml'))

    def test_parse_yaml_srclib_does_not_exists(self):
        with self.assertRaisesRegex(
            MetaDataException,
            "Invalid scrlib metadata: "
            r"'non(/|\\)existent-test-srclib.yml' "
            "does not exist",
        ):
            fdroidserver.metadata.parse_yaml_srclib(
                Path('non/existent-test-srclib.yml')
            )

    def test_parse_yaml_srclib_simple(self):
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            with Path('simple.yml').open('w', encoding='utf-8') as f:
                f.write(
                    textwrap.dedent(
                        '''\
                    # this should be simple
                    RepoType: git
                    Repo: https://git.host/repo.git
                    '''
                    )
                )
            srclib = fdroidserver.metadata.parse_yaml_srclib(Path('simple.yml'))
            self.assertDictEqual(
                {
                    'Repo': 'https://git.host/repo.git',
                    'RepoType': 'git',
                    'Subdir': None,
                    'Prepare': None,
                },
                srclib,
            )

    def test_parse_yaml_srclib_simple_with_blanks(self):
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            with Path('simple.yml').open('w', encoding='utf-8') as f:
                f.write(
                    textwrap.dedent(
                        '''\
                    # this should be simple

                    RepoType: git

                    Repo: https://git.host/repo.git

                    Subdir:

                    Prepare:
                    '''
                    )
                )
            srclib = fdroidserver.metadata.parse_yaml_srclib(Path('simple.yml'))
            self.assertDictEqual(
                {
                    'Repo': 'https://git.host/repo.git',
                    'RepoType': 'git',
                    'Subdir': [''],
                    'Prepare': [],
                },
                srclib,
            )

    def test_parse_yaml_srclib_Changelog_cketti(self):
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            with Path('Changelog-cketti.yml').open('w', encoding='utf-8') as f:
                f.write(
                    textwrap.dedent(
                        '''\
                    RepoType: git
                    Repo: https://github.com/cketti/ckChangeLog

                    Subdir: library,ckChangeLog/src/main
                    Prepare: "[ -f project.properties ] || echo 'source.dir=java' > ant.properties && echo -e 'android.library=true\\\\ntarget=android-19' > project.properties"
                    '''
                    )
                )
            srclib = fdroidserver.metadata.parse_yaml_srclib(
                Path('Changelog-cketti.yml')
            )
            self.assertDictEqual(
                srclib,
                {
                    'Repo': 'https://github.com/cketti/ckChangeLog',
                    'RepoType': 'git',
                    'Subdir': ['library', 'ckChangeLog/src/main'],
                    'Prepare': [
                        "[ -f project.properties ] || echo 'source.dir=java' > "
                        "ant.properties && echo -e "
                        "'android.library=true\\ntarget=android-19' > project.properties"
                    ],
                },
            )

    def test_read_srclibs_yml_subdir_list(self):
        fdroidserver.metadata.srclibs = None
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            Path('srclibs').mkdir()
            with Path('srclibs/with-list.yml').open('w', encoding='utf-8') as f:
                f.write(
                    textwrap.dedent(
                        '''\
                    # this should be simple
                    RepoType: git
                    Repo: https://git.host/repo.git

                    Subdir:
                     - This is your last chance.
                     - After this, there is no turning back.
                     - You take the blue pill—the story ends,
                     - you wake up in your bed
                     - and believe whatever you want to believe.
                     - You take the red pill—you stay in Wonderland
                     - and I show you how deep the rabbit-hole goes.
                    Prepare:
                        There is a difference between knowing the path
                        and walking the path.
                    '''
                    )
                )
            fdroidserver.metadata.read_srclibs()
        self.maxDiff = None
        self.assertDictEqual(
            fdroidserver.metadata.srclibs,
            {
                'with-list': {
                    'RepoType': 'git',
                    'Repo': 'https://git.host/repo.git',
                    'Subdir': [
                        'This is your last chance.',
                        'After this, there is no turning back.',
                        'You take the blue pill—the story ends,',
                        'you wake up in your bed',
                        'and believe whatever you want to believe.',
                        'You take the red pill—you stay in Wonderland',
                        'and I show you how deep the rabbit-hole goes.',
                    ],
                    'Prepare': [
                        'There is a difference between knowing the path '
                        'and walking the path.'
                    ],
                }
            },
        )

    def test_read_srclibs_yml_prepare_list(self):
        fdroidserver.metadata.srclibs = None
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            Path('srclibs').mkdir()
            with Path('srclibs/with-list.yml').open('w', encoding='utf-8') as f:
                f.write(
                    textwrap.dedent(
                        '''\
                    # this should be simple
                    RepoType: git
                    Repo: https://git.host/repo.git

                    Subdir:
                    Prepare:
                     - Many
                     - invalid
                     - commands
                     - here.
                    '''
                    )
                )
            fdroidserver.metadata.read_srclibs()
        self.maxDiff = None
        self.assertDictEqual(
            fdroidserver.metadata.srclibs,
            {
                'with-list': {
                    'RepoType': 'git',
                    'Repo': 'https://git.host/repo.git',
                    'Subdir': [''],
                    'Prepare': [
                        'Many',
                        'invalid',
                        'commands',
                        'here.',
                    ],
                }
            },
        )

    def test_read_srclibs(self):
        fdroidserver.metadata.srclibs = None
        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            Path('srclibs').mkdir()
            with Path('srclibs/simple.yml').open('w', encoding='utf-8') as f:
                f.write(
                    textwrap.dedent(
                        '''\
                    RepoType: git
                    Repo: https://git.host/repo.git
                    '''
                    )
                )
            with Path('srclibs/simple-wb.yml').open('w', encoding='utf-8') as f:
                f.write(
                    textwrap.dedent(
                        '''\
                    # this should be simple
                    RepoType: git
                    Repo: https://git.host/repo.git

                    Subdir:
                    Prepare:
                    '''
                    )
                )
            fdroidserver.metadata.read_srclibs()
        self.assertDictEqual(
            fdroidserver.metadata.srclibs,
            {
                'simple-wb': {
                    'RepoType': 'git',
                    'Repo': 'https://git.host/repo.git',
                    'Subdir': [''],
                    'Prepare': [],
                },
                'simple': {
                    'RepoType': 'git',
                    'Repo': 'https://git.host/repo.git',
                    'Subdir': None,
                    'Prepare': None,
                },
            },
        )

    def test_build_ndk_path(self):
        with tempfile.TemporaryDirectory(prefix='android-sdk-') as sdk_path:
            config = {'ndk_paths': {}, 'sdk_path': sdk_path}
            fdroidserver.common.config = config

            build = fdroidserver.metadata.Build()
            build.ndk = 'r10e'
            self.assertEqual('', build.ndk_path())

            correct = '/fake/path/ndk/r21b'
            config['ndk_paths'] = {'r21b': correct}
            self.assertEqual('', build.ndk_path())
            config['ndk_paths'] = {'r10e': correct}
            self.assertEqual(correct, build.ndk_path())

            r10e = '/fake/path/ndk/r10e'
            r22b = '/fake/path/ndk/r22e'
            config['ndk_paths'] = {'r10e': r10e, 'r22b': r22b}
            self.assertEqual(r10e, build.ndk_path())

            build.ndk = ['r10e', 'r22b']
            self.assertEqual(r10e, build.ndk_path())

            build.ndk = ['r22b', 'r10e']
            self.assertEqual(r22b, build.ndk_path())

    def test_build_ndk_path_only_accepts_str(self):
        """Paths in the config must be strings, never pathlib.Path instances"""
        config = {'ndk_paths': {'r24': Path('r24')}}
        fdroidserver.common.config = config
        build = fdroidserver.metadata.Build()
        build.ndk = 'r24'
        with self.assertRaises(TypeError):
            build.ndk_path()

    def test_del_duplicated_NoSourceSince(self):
        app = {
            'AntiFeatures': {'Ads': {}, 'NoSourceSince': {DEFAULT_LOCALE: '1.0'}},
            'NoSourceSince': '1.0',
        }
        metadata._del_duplicated_NoSourceSince(app)
        self.assertEqual(app, {'AntiFeatures': {'Ads': {}}, 'NoSourceSince': '1.0'})

    def test_check_manually_extended_NoSourceSince(self):
        app = {
            'AntiFeatures': {'NoSourceSince': {DEFAULT_LOCALE: '1.0', 'de': '1,0'}},
            'NoSourceSince': '1.0',
        }
        metadata._del_duplicated_NoSourceSince(app)
        self.assertEqual(
            app,
            {
                'AntiFeatures': {'NoSourceSince': {DEFAULT_LOCALE: '1.0', 'de': '1,0'}},
                'NoSourceSince': '1.0',
            },
        )

    def test_make_sure_nosourcesince_does_not_get_written(self):
        appid = 'com.politedroid'
        app = metadata.read_metadata({appid: -1})[appid]
        builds = app['Builds']
        app['Builds'] = [copy.deepcopy(builds[0])]
        mf = io.StringIO()
        metadata.write_yaml(mf, app)
        mf.seek(0)
        self.maxDiff = None
        self.assertEqual(
            mf.read(),
            textwrap.dedent(
                """\
                AntiFeatures:
                  - NonFreeNet
                Categories:
                  - Multimedia
                  - Security
                  - Time
                License: GPL-3.0-only
                SourceCode: https://github.com/miguelvps/PoliteDroid
                IssueTracker: https://github.com/miguelvps/PoliteDroid/issues

                AutoName: Polite Droid
                Summary: Calendar tool
                Description: Activates silent mode during calendar events.

                RepoType: git
                Repo: https://github.com/miguelvps/PoliteDroid.git

                Builds:
                  - versionName: '1.2'
                    versionCode: 3
                    commit: 6a548e4b19
                    target: android-10
                    antifeatures:
                      - KnownVuln
                      - NonFreeAssets
                      - UpstreamNonFree

                ArchivePolicy: 4
                AutoUpdateMode: Version v%v
                UpdateCheckMode: Tags
                CurrentVersion: '1.5'
                CurrentVersionCode: 6

                NoSourceSince: '1.5'
                """
            ),
        )

    def test_app_to_yaml_smokecheck(self):
        self.assertTrue(
            isinstance(metadata._app_to_yaml(dict()), ruamel.yaml.comments.CommentedMap)
        )

    def test_app_to_yaml_build_list_empty(self):
        app = metadata.App({'Builds': [metadata.Build({'rm': []})]})
        self.assertEqual(dict(), metadata._app_to_yaml(app)['Builds'][0])

    def test_app_to_yaml_build_list_one(self):
        app = metadata.App({'Builds': [metadata.Build({'rm': ['one']})]})
        self.assertEqual({'rm': ['one']}, metadata._app_to_yaml(app)['Builds'][0])

    def test_app_to_yaml_build_list_two(self):
        app = metadata.App({'Builds': [metadata.Build({'rm': ['1', '2']})]})
        self.assertEqual({'rm': ['1', '2']}, metadata._app_to_yaml(app)['Builds'][0])

    def test_app_to_yaml_build_list(self):
        app = metadata.App({'Builds': [metadata.Build({'rm': ['b2', 'NO1']})]})
        self.assertEqual({'rm': ['b2', 'NO1']}, metadata._app_to_yaml(app)['Builds'][0])

    def test_app_to_yaml_AllowedAPKSigningKeys_two(self):
        cm = metadata._app_to_yaml(metadata.App({'AllowedAPKSigningKeys': ['b', 'A']}))
        self.assertEqual(['b', 'a'], cm['AllowedAPKSigningKeys'])

    def test_app_to_yaml_AllowedAPKSigningKeys_one(self):
        cm = metadata._app_to_yaml(metadata.App({'AllowedAPKSigningKeys': ['One']}))
        self.assertEqual('one', cm['AllowedAPKSigningKeys'])

    def test_app_to_yaml_int_hex(self):
        cm = metadata._app_to_yaml(metadata.App({'CurrentVersionCode': 0xFF}))
        self.assertEqual(255, cm['CurrentVersionCode'])

    def test_app_to_yaml_int_underscore(self):
        cm = metadata._app_to_yaml(metadata.App({'CurrentVersionCode': 1_2_3}))
        self.assertEqual(123, cm['CurrentVersionCode'])

    def test_app_to_yaml_int_0(self):
        """Document that 0 values fail to make it through."""
        # TODO it should be possible to use `CurrentVersionCode: 0`
        cm = metadata._app_to_yaml(metadata.App({'CurrentVersionCode': 0}))
        self.assertFalse('CurrentVersionCode' in cm)

    def test_format_multiline(self):
        self.assertEqual(metadata._format_multiline('description'), 'description')

    def test_format_multiline_empty(self):
        self.assertEqual(metadata._format_multiline(''), '')

    def test_format_multiline_newline_char(self):
        self.assertEqual(metadata._format_multiline('one\\ntwo'), 'one\\ntwo')

    def test_format_multiline_newlines(self):
        self.assertEqual(
            metadata._format_multiline(
                textwrap.dedent(
                    """
                    one
                    two
                    three
                    """
                )
            ),
            '\none\ntwo\nthree\n',
        )

    def test_format_list_empty(self):
        self.assertEqual(metadata._format_list(['', None]), list())

    def test_format_list_one_empty(self):
        self.assertEqual(metadata._format_list(['foo', None]), ['foo'])

    def test_format_list_two(self):
        self.assertEqual(metadata._format_list(['2', '1']), ['2', '1'])

    def test_format_list_newline(self):
        self.assertEqual(metadata._format_list(['one\ntwo']), ['one\ntwo'])

    def test_format_list_newline_char(self):
        self.assertEqual(metadata._format_list(['one\\ntwo']), ['one\\ntwo'])

    def test_format_script_empty(self):
        self.assertEqual(metadata._format_script(['', None]), list())

    def test_format_script_newline(self):
        self.assertEqual(metadata._format_script(['one\ntwo']), 'one\ntwo')

    def test_format_script_newline_char(self):
        self.assertEqual(metadata._format_script(['one\\ntwo']), 'one\\ntwo')

    def test_format_stringmap_empty(self):
        self.assertEqual(
            metadata._format_stringmap('🔥', 'test', dict()),
            list(),
        )

    def test_format_stringmap_one_list(self):
        self.assertEqual(
            metadata._format_stringmap('🔥', 'test', {'Tracking': {}, 'Ads': {}}),
            ['Ads', 'Tracking'],
        )

    def test_format_stringmap_one_list_empty_desc(self):
        self.assertEqual(
            metadata._format_stringmap('🔥', 'test', {'NonFree': {}, 'Ads': {'en': ''}}),
            ['Ads', 'NonFree'],
        )

    def test_format_stringmap_three_list(self):
        self.assertEqual(
            metadata._format_stringmap('🔥', 'test', {'B': {}, 'A': {}, 'C': {}}),
            ['A', 'B', 'C'],
        )

    def test_format_stringmap_two_dict(self):
        self.assertEqual(
            metadata._format_stringmap('🔥', 'test', {'1': {'uz': 'a'}, '2': {}}),
            {'1': {'uz': 'a'}, '2': {}},
        )

    def test_format_stringmap_three_locales(self):
        self.assertEqual(
            metadata._format_stringmap(
                '🔥', 'test', {'AF': {'uz': 'a', 'ko': 'b', 'zh': 'c'}}
            ),
            {'AF': {'ko': 'b', 'uz': 'a', 'zh': 'c'}},
        )

    def test_format_stringmap_move_build_antifeatures_to_filesystem(self):
        os.chdir(self.testdir)
        appid = 'a'
        yml = Path('metadata/a.yml')
        yml.parent.mkdir()
        self.assertEqual(
            metadata._format_stringmap(
                appid, 'antifeatures', {'AF': {'uz': 'a', 'ko': 'b', 'zh': 'c'}}
            ),
            {'AF': {'ko': 'b', 'uz': 'a', 'zh': 'c'}},
        )

    def test_format_stringmap_app_antifeatures_conflict(self):
        """Raise an error if a YAML Anti-Feature conflicts with a localized file."""
        os.chdir(self.testdir)
        appid = 'a'
        field = 'AntiFeatures'
        locale = 'ko'
        yml = Path('metadata/a.yml')
        antifeatures_ko = yml.parent / appid / locale / field.lower()
        antifeatures_ko.mkdir(parents=True)
        afname = 'Anti-🔥'
        (antifeatures_ko / (afname + '.txt')).write_text('SOMETHING ELSE')
        with self.assertRaises(MetaDataException):
            metadata._format_stringmap(
                appid, field, {afname: {'uz': 'a', locale: 'b', 'zh': 'c'}}
            )

    def test_format_stringmap_app_antifeatures_conflict_same_contents(self):
        """Raise an error if a YAML Anti-Feature conflicts with a localized file."""
        os.chdir(self.testdir)
        appid = 'a'
        field = 'AntiFeatures'
        locale = 'ko'
        yml = Path('metadata/a.yml')
        antifeatures_ko = yml.parent / appid / locale / field.lower()
        antifeatures_ko.mkdir(parents=True)
        afname = 'Anti-🔥'
        (antifeatures_ko / (afname + '.txt')).write_text('b')
        metadata._format_stringmap(
            appid, field, {afname: {'uz': 'a', locale: 'b', 'zh': 'c'}}
        )

    def test_format_stringmap_build_antifeatures_conflict(self):
        """Raise an error if a YAML Anti-Feature conflicts with a localized file."""
        os.chdir(self.testdir)
        appid = 'a'
        field = 'antifeatures'
        locale = 'ko'
        versionCode = 123
        yml = Path('metadata/a.yml')
        antifeatures_ko = yml.parent / appid / locale / field.lower()
        antifeatures_ko.mkdir(parents=True)
        afname = 'Anti-🔥'
        with (antifeatures_ko / ('%d_%s.txt' % (versionCode, afname))).open('w') as fp:
            fp.write('SOMETHING ELSE')
        with self.assertRaises(MetaDataException):
            metadata._format_stringmap(
                appid, field, {afname: {'uz': 'a', locale: 'b', 'zh': 'c'}}, versionCode
            )

    def test_app_to_yaml_one_category(self):
        """Categories does not get simplified to string when outputting YAML."""
        self.assertEqual(
            metadata._app_to_yaml({'Categories': ['one']}),
            {'Categories': ['one']},
        )

    def test_app_to_yaml_categories(self):
        """Sort case-insensitive before outputting YAML."""
        self.assertEqual(
            metadata._app_to_yaml({'Categories': ['c', 'a', 'B']}),
            {'Categories': ['a', 'B', 'c']},
        )

    def test_builds_to_yaml_gradle_yes(self):
        app = {'Builds': [{'versionCode': 0, 'gradle': ['yes']}]}
        self.assertEqual(
            metadata._builds_to_yaml(app), [{'versionCode': 0, 'gradle': ['yes']}]
        )

    def test_builds_to_yaml_gradle_off(self):
        app = {'Builds': [{'versionCode': 0, 'gradle': ['off']}]}
        self.assertEqual(
            metadata._builds_to_yaml(app), [{'versionCode': 0, 'gradle': ['off']}]
        )

    def test_builds_to_yaml_gradle_true(self):
        app = {'Builds': [{'versionCode': 0, 'gradle': ['true']}]}
        self.assertEqual(
            metadata._builds_to_yaml(app), [{'versionCode': 0, 'gradle': ['true']}]
        )

    def test_builds_to_yaml_gradle_false(self):
        app = {'Builds': [{'versionCode': 0, 'gradle': ['false']}]}
        self.assertEqual(
            metadata._builds_to_yaml(app), [{'versionCode': 0, 'gradle': ['false']}]
        )

    def test_builds_to_yaml_stripped(self):
        self.assertEqual(
            metadata._builds_to_yaml(
                {
                    'Builds': [
                        metadata.Build({'versionCode': 0, 'rm': [None], 'init': ['']})
                    ]
                }
            ),
            [{'versionCode': 0}],
        )

    def test_builds_to_yaml(self):
        """Include one of each flag type with a valid value."""
        app = {
            'Builds': [
                metadata.Build(
                    {
                        'versionCode': 0,
                        'gradle': ['free'],
                        'rm': ['0', '2'],
                        'submodules': True,
                        'timeout': 0,
                        'init': ['false', 'two'],
                    }
                )
            ]
        }
        # check that metadata.Build() inited flag values
        self.assertEqual(app['Builds'][0]['scanignore'], list())
        # then unchanged values should be removed by _builds_to_yaml
        self.assertEqual(
            metadata._builds_to_yaml(app),
            [
                {
                    'versionCode': 0,
                    'gradle': ['free'],
                    'rm': ['0', '2'],
                    'submodules': True,
                    'timeout': 0,
                    'init': ['false', 'two'],
                }
            ],
        )


class PostMetadataParseTest(unittest.TestCase):
    """Test the functions that post process the YAML input.

    The following series of "post_metadata_parse" tests map out the
    current state of automatic type conversion in the YAML post
    processing.  They are not necessary a statement of how things
    should be, but more to surface the details of it functions.

    """

    def setUp(self):
        fdroidserver.metadata.warnings_action = 'error'

    def _post_metadata_parse_app_int(self, from_yaml, expected):
        app = {'ArchivePolicy': from_yaml}
        metadata.post_parse_yaml_metadata(app)
        return {'ArchivePolicy': expected}, app

    def _post_metadata_parse_app_list(self, from_yaml, expected):
        app = {'AllowedAPKSigningKeys': from_yaml}
        metadata.post_parse_yaml_metadata(app)
        return {'AllowedAPKSigningKeys': expected}, app

    def _post_metadata_parse_app_string(self, from_yaml, expected):
        app = {'Repo': from_yaml}
        metadata.post_parse_yaml_metadata(app)
        return {'Repo': expected}, app

    def _post_metadata_parse_build_bool(self, from_yaml, expected):
        tested_key = 'submodules'
        app = {'Builds': [{'versionCode': 1, tested_key: from_yaml}]}
        post = copy.deepcopy(app)
        metadata.post_parse_yaml_metadata(post)
        del app['Builds'][0]['versionCode']
        del post['Builds'][0]['versionCode']
        for build in post['Builds']:
            for k in list(build):
                if k != tested_key:
                    del build[k]
        app['Builds'][0][tested_key] = expected
        return app, post

    def _post_metadata_parse_build_int(self, from_yaml, expected):
        tested_key = 'versionCode'
        app = {'Builds': [{'versionCode': from_yaml}]}
        post = copy.deepcopy(app)
        metadata.post_parse_yaml_metadata(post)
        for build in post['Builds']:
            for k in list(build):
                if k != tested_key:
                    del build[k]
        app['Builds'][0][tested_key] = expected
        return app, post

    def _post_metadata_parse_build_list(self, from_yaml, expected):
        tested_key = 'rm'
        app = {'Builds': [{'versionCode': 1, tested_key: from_yaml}]}
        post = copy.deepcopy(app)
        metadata.post_parse_yaml_metadata(post)
        del app['Builds'][0]['versionCode']
        del post['Builds'][0]['versionCode']
        for build in post['Builds']:
            for k in list(build):
                if k != tested_key:
                    del build[k]
        app['Builds'][0][tested_key] = expected
        return app, post

    def _post_metadata_parse_build_script(self, from_yaml, expected):
        tested_key = 'build'
        app = {'Builds': [{'versionCode': 1, tested_key: from_yaml}]}
        post = copy.deepcopy(app)
        metadata.post_parse_yaml_metadata(post)
        del app['Builds'][0]['versionCode']
        del post['Builds'][0]['versionCode']
        for build in post['Builds']:
            for k in list(build):
                if k != tested_key:
                    del build[k]
        app['Builds'][0][tested_key] = expected
        return app, post

    def _post_metadata_parse_build_string(self, from_yaml, expected):
        tested_key = 'commit'
        app = {'Builds': [{'versionCode': 1, tested_key: from_yaml}]}
        post = copy.deepcopy(app)
        metadata.post_parse_yaml_metadata(post)
        del app['Builds'][0]['versionCode']
        del post['Builds'][0]['versionCode']
        for build in post['Builds']:
            for k in list(build):
                if k != tested_key:
                    del build[k]
        app['Builds'][0][tested_key] = expected
        return app, post

    def test_post_metadata_parse_none(self):
        """Run None aka YAML null or blank through the various field and flag types."""
        self.assertEqual(*self._post_metadata_parse_app_int(None, None))
        self.assertEqual(*self._post_metadata_parse_app_list(None, None))
        self.assertEqual(*self._post_metadata_parse_app_string(None, None))
        self.assertEqual(*self._post_metadata_parse_build_bool(None, None))
        self.assertEqual(*self._post_metadata_parse_build_int(None, None))
        self.assertEqual(*self._post_metadata_parse_build_list(None, None))
        self.assertEqual(*self._post_metadata_parse_build_script(None, None))
        self.assertEqual(*self._post_metadata_parse_build_string(None, None))

    def test_post_metadata_parse_int(self):
        """Run the int 123456 through the various field and flag types."""
        self.assertEqual(*self._post_metadata_parse_app_int(123456, 123456))
        self.assertEqual(*self._post_metadata_parse_app_list(123456, ['123456']))
        self.assertEqual(*self._post_metadata_parse_app_string(123456, '123456'))
        self.assertEqual(*self._post_metadata_parse_build_bool(123456, True))
        self.assertEqual(*self._post_metadata_parse_build_int(123456, 123456))
        self.assertEqual(*self._post_metadata_parse_build_list(123456, ['123456']))
        self.assertEqual(*self._post_metadata_parse_build_script(123456, ['123456']))
        self.assertEqual(*self._post_metadata_parse_build_string(123456, '123456'))

    def test_post_metadata_parse_sha256(self):
        """Run a SHA-256 that YAML calls an int through the various types.

        The current f-droid.org signer set has SHA-256 values with a
        maximum of two leading zeros, but this will handle more.

        """
        yaml = ruamel.yaml.YAML(typ='safe', pure=True)
        str_sha256 = '0000000000000498456908409534729834729834729834792837487293847926'
        sha256 = yaml.load('a: ' + str_sha256)['a']
        self.assertEqual(*self._post_metadata_parse_app_int(sha256, int(str_sha256)))
        self.assertEqual(*self._post_metadata_parse_app_list(sha256, [str_sha256]))
        self.assertEqual(*self._post_metadata_parse_app_string(sha256, str_sha256))
        self.assertEqual(*self._post_metadata_parse_build_bool(sha256, True))
        self.assertEqual(*self._post_metadata_parse_build_int(sha256, sha256))
        self.assertEqual(*self._post_metadata_parse_build_list(sha256, [str_sha256]))
        self.assertEqual(*self._post_metadata_parse_build_script(sha256, [str_sha256]))
        self.assertEqual(*self._post_metadata_parse_build_string(sha256, str_sha256))

    def test_post_metadata_parse_int_0(self):
        """Run the int 0 through the various field and flag types."""
        self.assertEqual(*self._post_metadata_parse_app_int(0, 0))
        self.assertEqual(*self._post_metadata_parse_app_list(0, ['0']))
        self.assertEqual(*self._post_metadata_parse_app_string(0, '0'))
        self.assertEqual(*self._post_metadata_parse_build_bool(0, False))
        self.assertEqual(*self._post_metadata_parse_build_int(0, 0))
        self.assertEqual(*self._post_metadata_parse_build_list(0, ['0']))
        self.assertEqual(*self._post_metadata_parse_build_script(0, ['0']))
        self.assertEqual(*self._post_metadata_parse_build_string(0, '0'))

    def test_post_metadata_parse_float_0_0(self):
        """Run the float 0.0 through the various field and flag types."""
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_app_int(0.0, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_app_list(0.0, ['0.0']))
        self.assertEqual(*self._post_metadata_parse_app_string(0.0, '0.0'))
        self.assertEqual(*self._post_metadata_parse_build_bool(0.0, False))
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_build_int(0.0, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_build_list(0.0, ['0.0']))
        self.assertEqual(*self._post_metadata_parse_build_script(0.0, ['0.0']))
        self.assertEqual(*self._post_metadata_parse_build_string(0.0, '0.0'))

    def test_post_metadata_parse_float_0_1(self):
        """Run the float 0.1 through the various field and flag types."""
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_app_int(0.1, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_app_list(0.1, ['0.1']))
        self.assertEqual(*self._post_metadata_parse_app_string(0.1, '0.1'))
        self.assertEqual(*self._post_metadata_parse_build_bool(0.1, True))
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_build_int(0.1, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_build_list(0.1, ['0.1']))
        self.assertEqual(*self._post_metadata_parse_build_script(0.1, ['0.1']))
        self.assertEqual(*self._post_metadata_parse_build_string(0.1, '0.1'))

    def test_post_metadata_parse_float_1_0(self):
        """Run the float 1.0 through the various field and flag types."""
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_app_int(1.0, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_app_list(1.0, ['1.0']))
        self.assertEqual(*self._post_metadata_parse_app_string(1.0, '1.0'))
        self.assertEqual(*self._post_metadata_parse_build_bool(1.0, True))
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_build_int(1.0, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_build_list(1.0, ['1.0']))
        self.assertEqual(*self._post_metadata_parse_build_script(1.0, ['1.0']))
        self.assertEqual(*self._post_metadata_parse_build_string(1.0, '1.0'))

    def test_post_metadata_parse_empty_list(self):
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_app_int(list(), MetaDataException)
        self.assertEqual(*self._post_metadata_parse_app_list(list(), list()))
        self.assertEqual(*self._post_metadata_parse_app_string(list(), list()))
        self.assertEqual(*self._post_metadata_parse_build_bool(list(), False))
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_build_int(list(), MetaDataException)
        self.assertEqual(*self._post_metadata_parse_build_list(list(), list()))
        self.assertEqual(*self._post_metadata_parse_build_script(list(), list()))
        self.assertEqual(*self._post_metadata_parse_build_string(list(), list()))

    def test_post_metadata_parse_set_of_1(self):
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_app_int({1}, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_app_list({1}, ['1']))
        self.assertEqual(*self._post_metadata_parse_app_string({1}, '{1}'))
        self.assertEqual(*self._post_metadata_parse_build_bool({1}, True))
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_build_int({1}, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_build_list({1}, ['1']))
        self.assertEqual(*self._post_metadata_parse_build_script({1}, ['1']))
        self.assertEqual(*self._post_metadata_parse_build_string({1}, '{1}'))

    def test_post_metadata_parse_empty_dict(self):
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_app_int(dict(), MetaDataException)
        self.assertEqual(*self._post_metadata_parse_app_list(dict(), dict()))
        self.assertEqual(*self._post_metadata_parse_app_string(dict(), dict()))
        self.assertEqual(*self._post_metadata_parse_build_bool(dict(), False))
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_build_int(dict(), MetaDataException)
        self.assertEqual(*self._post_metadata_parse_build_list(dict(), dict()))
        self.assertEqual(*self._post_metadata_parse_build_script(dict(), dict()))
        self.assertEqual(*self._post_metadata_parse_build_string(dict(), dict()))

    def test_post_metadata_parse_list_int_string(self):
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_app_int([1, 'a'], MetaDataException)
        self.assertEqual(*self._post_metadata_parse_app_list([1, 'a'], ['1', 'a']))
        self.assertEqual(*self._post_metadata_parse_app_string([1, 'a'], "[1, 'a']"))
        self.assertEqual(*self._post_metadata_parse_build_bool([1, 'a'], True))
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_build_int([1, 'a'], MetaDataException)
        self.assertEqual(*self._post_metadata_parse_build_list([1, 'a'], ['1', 'a']))
        self.assertEqual(*self._post_metadata_parse_build_script([1, 'a'], ['1', 'a']))
        self.assertEqual(*self._post_metadata_parse_build_string([1, 'a'], "[1, 'a']"))

    def test_post_metadata_parse_dict_int_string(self):
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_app_int({'k': 1}, MetaDataException)
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_app_list({'k': 1}, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_app_string({'k': 1}, "{'k': 1}"))
        self.assertEqual(*self._post_metadata_parse_build_bool({'k': 1}, True))
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_build_int({'k': 1}, MetaDataException)
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_build_list({'k': 1}, MetaDataException)
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_build_script({'k': 1}, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_build_string({'k': 1}, "{'k': 1}"))

    def test_post_metadata_parse_false(self):
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_app_int(False, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_app_list(False, ['false']))
        self.assertEqual(*self._post_metadata_parse_app_string(False, 'false'))
        self.assertEqual(*self._post_metadata_parse_build_bool(False, False))
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_build_int(False, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_build_list(False, ['false']))
        self.assertEqual(*self._post_metadata_parse_build_script(False, ['false']))
        self.assertEqual(*self._post_metadata_parse_build_string(False, 'false'))

    def test_post_metadata_parse_true(self):
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_app_int(True, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_app_list(True, ['true']))
        self.assertEqual(*self._post_metadata_parse_app_string(True, 'true'))
        self.assertEqual(*self._post_metadata_parse_build_bool(True, True))
        with self.assertRaises(MetaDataException):
            self._post_metadata_parse_build_int(True, MetaDataException)
        self.assertEqual(*self._post_metadata_parse_build_list(True, ['true']))
        self.assertEqual(*self._post_metadata_parse_build_script(True, ['true']))
        self.assertEqual(*self._post_metadata_parse_build_string(True, 'true'))




# TODO convert to config.yml!

repo_url = "https://MyFirstFDroidRepo.org/fdroid/repo"
repo_name = "My First F-Droid Repo Demo"
repo_description = """This is a repository of apps to be used with F-Droid. Applications in this repository are either official binaries built by the original application developers, or are binaries built from source by the admin of f-droid.org using the tools on https://gitlab.com/fdroid."""

archive_older = 3
archive_url = "https://f-droid.org/archive"
archive_name = "My First F-Droid Archive Demo"
archive_description = """
The repository of older versions of applications from the main demo repository.
"""

make_current_version_link = False

repo_keyalias = "sova"
keystore = "keystore.jks"
keystorepass = "r9aquRHYoI8+dYz6jKrLntQ5/NJNASFBacJh7Jv2BlI="
keypass = "r9aquRHYoI8+dYz6jKrLntQ5/NJNASFBacJh7Jv2BlI="
keydname = "CN=sova, OU=F-Droid"

mirrors = (
    'http://foobarfoobarfoobar.onion/fdroid',
    'https://foo.bar/fdroid',
)

install_list = 'org.adaway'
uninstall_list = ('com.android.vending', 'com.facebook.orca', )

repo_key_sha256 = "f49af3f11efddf20dffd70f5e3117b9976674167adca280e6b1932a0601b26f6"



#!/usr/bin/env python3
"""Set up an app build for a nightly build repo."""
#
# nightly.py - part of the FDroid server tools
# Copyright (C) 2017 Hans-Christoph Steiner <hans@eds.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import datetime
import git
import hashlib
import logging
import os
import paramiko
import platform
import shutil
import ssl
import subprocess
import sys
import tempfile
import yaml
from urllib.parse import urlparse
from argparse import ArgumentParser
from typing import Optional

from . import _
from . import common
from .exception import VCSException

# hard coded defaults for Android ~/.android/debug.keystore files
# https://developers.google.com/android/guides/client-auth
KEYSTORE_FILE = os.path.join(os.getenv('HOME'), '.android', 'debug.keystore')
PASSWORD = 'android'  # nosec B105 standard hardcoded password for debug keystores
KEY_ALIAS = 'androiddebugkey'
DISTINGUISHED_NAME = 'CN=Android Debug,O=Android,C=US'

# standard suffix for naming fdroid git repos
NIGHTLY = '-nightly'


def _get_keystore_secret_var(keystore: str) -> str:
    """Get keystore secret as base64.

    Parameters
    ----------
    keystore
        The path of the keystore.

    Returns
    -------
    base64_secret
        The keystore secret as base64 string.
    """
    with open(keystore, 'rb') as fp:
        return base64.standard_b64encode(fp.read()).decode('ascii')


def _ssh_key_from_debug_keystore(keystore: Optional[str] = None) -> str:
    """Convert a debug keystore to an SSH private key.

    This leaves the original keystore file in place.

    Parameters
    ----------
    keystore
        The keystore to convert to a SSH private key.

    Returns
    -------
    key_path
        The SSH private key file path in the temporary directory.
    """
    if keystore is None:
        # set this here so it can be overridden in the tests
        # TODO convert this to a class to get rid of this nonsense
        keystore = KEYSTORE_FILE
    tmp_dir = tempfile.mkdtemp(prefix='.')
    privkey = os.path.join(tmp_dir, '.privkey')
    key_pem = os.path.join(tmp_dir, '.key.pem')
    p12 = os.path.join(tmp_dir, '.keystore.p12')
    _config = dict()
    common.fill_config_defaults(_config)
    subprocess.check_call(
        [
            _config['keytool'],
            '-importkeystore',
            '-srckeystore',
            keystore,
            '-srcalias',
            KEY_ALIAS,
            '-srcstorepass',
            PASSWORD,
            '-srckeypass',
            PASSWORD,
            '-destkeystore',
            p12,
            '-destalias',
            KEY_ALIAS,
            '-deststorepass',
            PASSWORD,
            '-destkeypass',
            PASSWORD,
            '-deststoretype',
            'PKCS12',
        ],
        env={'LC_ALL': 'C.UTF-8'},
    )
    subprocess.check_call(
        [
            'openssl',
            'pkcs12',
            '-in',
            p12,
            '-out',
            key_pem,
            '-passin',
            'pass:' + PASSWORD,
            '-passout',
            'pass:' + PASSWORD,
        ],
        env={'LC_ALL': 'C.UTF-8'},
    )

    # OpenSSL 3.0 changed the default output format from PKCS#1 to
    # PKCS#8, which paramiko does not support.
    # https://www.openssl.org/docs/man3.0/man1/openssl-rsa.html#traditional
    # https://github.com/paramiko/paramiko/issues/1015
    openssl_rsa_cmd = ['openssl', 'rsa']
    if ssl.OPENSSL_VERSION_INFO[0] >= 3:
        openssl_rsa_cmd += ['-traditional']
    subprocess.check_call(
        openssl_rsa_cmd
        + [
            '-in',
            key_pem,
            '-out',
            privkey,
            '-passin',
            'pass:' + PASSWORD,
        ],
        env={'LC_ALL': 'C.UTF-8'},
    )
    os.remove(key_pem)
    os.remove(p12)
    os.chmod(privkey, 0o600)  # os.umask() should cover this, but just in case

    rsakey = paramiko.RSAKey.from_private_key_file(privkey)
    fingerprint = (
        base64.b64encode(hashlib.sha256(rsakey.asbytes()).digest())
        .decode('ascii')
        .rstrip('=')
    )
    ssh_private_key_file = os.path.join(
        tmp_dir, 'debug_keystore_' + fingerprint.replace('/', '_') + '_id_rsa'
    )
    shutil.move(privkey, ssh_private_key_file)

    pub = rsakey.get_name() + ' ' + rsakey.get_base64() + ' ' + ssh_private_key_file
    with open(ssh_private_key_file + '.pub', 'w') as fp:
        fp.write(pub)

    logging.info(_('\nSSH public key to be used as deploy key:') + '\n' + pub)

    return ssh_private_key_file


def get_repo_base_url(clone_url: str, repo_git_base: str, force_type: Optional[str] = None) -> str:
    """Generate the base URL for the F-Droid repository.

    Parameters
    ----------
    clone_url
        The URL to clone the Git repository.
    repo_git_base
        The project path of the Git repository at the Git forge.
    force_type
        The Git forge of the project.

    Returns
    -------
    repo_base_url
        The base URL of the F-Droid repository.
    """
    if force_type is None:
        force_type = urlparse(clone_url).netloc
    if force_type == 'gitlab.com':
        return clone_url + '/-/raw/master/fdroid'
    if force_type == 'github.com':
        return 'https://raw.githubusercontent.com/%s/master/fdroid' % repo_git_base
    print(_('ERROR: unsupported git host "%s", patches welcome!') % force_type)
    sys.exit(1)


def main():
    """Deploy to F-Droid repository or generate SSH private key from keystore.

    The behaviour of this function is influenced by the configuration file as
    well as command line parameters.

    Raises
    ------
    :exc:`~fdroidserver.exception.VCSException`
        If the nightly Git repository could not be cloned during an attempt to
        deploy.
    """
    parser = ArgumentParser()
    common.setup_global_opts(parser)
    parser.add_argument(
        "--keystore",
        default=KEYSTORE_FILE,
        help=_("Specify which debug keystore file to use."),
    )
    parser.add_argument(
        "--show-secret-var",
        action="store_true",
        default=False,
        help=_("Print the secret variable to the terminal for easy copy/paste"),
    )
    parser.add_argument(
        "--keep-private-keys",
        action="store_true",
        default=False,
        help=_("Do not remove the private keys generated from the keystore"),
    )
    parser.add_argument(
        "--no-deploy",
        action="store_true",
        default=False,
        help=_("Do not deploy the new files to the repo"),
    )
    parser.add_argument(
        "--file",
        default='app/build/outputs/apk/*.apk',
        help=_('The file to be included in the repo (path or glob)'),
    )
    parser.add_argument(
        "--no-checksum",
        action="store_true",
        default=False,
        help=_("Don't use rsync checksums"),
    )
    archive_older_unset = -1
    parser.add_argument(
        "--archive-older",
        type=int,
        default=archive_older_unset,
        help=_("Set maximum releases in repo before older ones are archived"),
    )
    # TODO add --with-btlog
    options = common.parse_args(parser)

    # force a tighter umask since this writes private key material
    umask = os.umask(0o077)

    if 'CI' in os.environ:
        v = os.getenv('DEBUG_KEYSTORE')
        debug_keystore = None
        if v:
            debug_keystore = base64.b64decode(v)
        if not debug_keystore:
            logging.error(_('DEBUG_KEYSTORE is not set or the value is incomplete'))
            sys.exit(1)
        os.makedirs(os.path.dirname(KEYSTORE_FILE), exist_ok=True)
        if os.path.exists(KEYSTORE_FILE):
            logging.warning(_('overwriting existing {path}').format(path=KEYSTORE_FILE))
        with open(KEYSTORE_FILE, 'wb') as fp:
            fp.write(debug_keystore)

        repo_basedir = os.path.join(os.getcwd(), 'fdroid')
        repodir = os.path.join(repo_basedir, 'repo')
        cibase = os.getcwd()
        os.makedirs(repodir, exist_ok=True)

        # the 'master' branch is hardcoded in fdroidserver/deploy.py
        if 'CI_PROJECT_PATH' in os.environ and 'CI_PROJECT_URL' in os.environ:
            # we are in GitLab CI
            repo_git_base = os.getenv('CI_PROJECT_PATH') + NIGHTLY
            clone_url = os.getenv('CI_PROJECT_URL') + NIGHTLY
            repo_base = get_repo_base_url(clone_url, repo_git_base, force_type='gitlab.com')
            servergitmirror = 'git@' + urlparse(clone_url).netloc + ':' + repo_git_base
            deploy_key_url = clone_url + '/-/settings/repository#js-deploy-keys-settings'
            git_user_name = os.getenv('GITLAB_USER_NAME')
            git_user_email = os.getenv('GITLAB_USER_EMAIL')
        elif 'TRAVIS_REPO_SLUG' in os.environ:
            # we are in Travis CI
            repo_git_base = os.getenv('TRAVIS_REPO_SLUG') + NIGHTLY
            clone_url = 'https://github.com/' + repo_git_base
            repo_base = get_repo_base_url(clone_url, repo_git_base, force_type='github.com')
            servergitmirror = 'git@github.com:' + repo_git_base
            deploy_key_url = ('https://github.com/' + repo_git_base + '/settings/keys'
                              + '\nhttps://developer.github.com/v3/guides/managing-deploy-keys/#deploy-keys')
            git_user_name = repo_git_base
            git_user_email = os.getenv('USER') + '@' + platform.node()
        elif (
            'CIRCLE_REPOSITORY_URL' in os.environ
            and 'CIRCLE_PROJECT_USERNAME' in os.environ
            and 'CIRCLE_PROJECT_REPONAME' in os.environ
        ):
            # we are in Circle CI
            repo_git_base = (os.getenv('CIRCLE_PROJECT_USERNAME')
                             + '/' + os.getenv('CIRCLE_PROJECT_REPONAME') + NIGHTLY)
            clone_url = os.getenv('CIRCLE_REPOSITORY_URL') + NIGHTLY
            repo_base = get_repo_base_url(clone_url, repo_git_base, force_type='github.com')
            servergitmirror = 'git@' + urlparse(clone_url).netloc + ':' + repo_git_base
            deploy_key_url = ('https://github.com/' + repo_git_base + '/settings/keys'
                              + '\nhttps://developer.github.com/v3/guides/managing-deploy-keys/#deploy-keys')
            git_user_name = os.getenv('CIRCLE_USERNAME')
            git_user_email = git_user_name + '@' + platform.node()
        elif 'GITHUB_ACTIONS' in os.environ:
            # we are in Github actions
            repo_git_base = (os.getenv('GITHUB_REPOSITORY') + NIGHTLY)
            clone_url = (os.getenv('GITHUB_SERVER_URL') + '/' + repo_git_base)
            repo_base = get_repo_base_url(clone_url, repo_git_base, force_type='github.com')
            servergitmirror = 'git@' + urlparse(clone_url).netloc + ':' + repo_git_base
            deploy_key_url = ('https://github.com/' + repo_git_base + '/settings/keys'
                              + '\nhttps://developer.github.com/v3/guides/managing-deploy-keys/#deploy-keys')
            git_user_name = os.getenv('GITHUB_ACTOR')
            git_user_email = git_user_name + '@' + platform.node()
        else:
            print(_('ERROR: unsupported CI type, patches welcome!'))
            sys.exit(1)

        repo_url = repo_base + '/repo'
        git_mirror_path = os.path.join(repo_basedir, 'git-mirror')
        git_mirror_fdroiddir = os.path.join(git_mirror_path, 'fdroid')
        git_mirror_repodir = os.path.join(git_mirror_fdroiddir, 'repo')
        git_mirror_metadatadir = os.path.join(git_mirror_fdroiddir, 'metadata')
        git_mirror_statsdir = os.path.join(git_mirror_fdroiddir, 'stats')
        if not os.path.isdir(git_mirror_repodir):
            logging.debug(_('cloning {url}').format(url=clone_url))
            vcs = common.getvcs('git', clone_url, git_mirror_path)
            p = vcs.git(['clone', '--', vcs.remote, str(vcs.local)])
            if p.returncode != 0:
                print('WARNING: only public git repos are supported!')
                raise VCSException('git clone %s failed:' % clone_url, p.output)
        if not os.path.isdir(git_mirror_repodir):
            os.makedirs(git_mirror_repodir, mode=0o755)

        mirror_git_repo = git.Repo.init(git_mirror_path)
        writer = mirror_git_repo.config_writer()
        writer.set_value('user', 'name', git_user_name)
        writer.set_value('user', 'email', git_user_email)
        writer.release()
        for remote in mirror_git_repo.remotes:
            mirror_git_repo.delete_remote(remote)

        readme_path = os.path.join(git_mirror_path, 'README.md')
        readme = '''
# {repo_git_base}

This is an app repository for nightly versions.
You can use it with the [F-Droid](https://f-droid.org/) Android app.

[![{repo_url}]({repo_url}/icons/icon.png)](https://fdroid.link/#{repo_url})

Last updated: {date}'''.format(repo_git_base=repo_git_base,
                               repo_url=repo_url,
                               date=datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S UTC'))
        with open(readme_path, 'w') as fp:
            fp.write(readme)
        mirror_git_repo.git.add(all=True)
        mirror_git_repo.index.commit("update README")

        mirror_git_repo.git.add(all=True)
        mirror_git_repo.index.commit("update repo/website icon")

        os.chdir(repo_basedir)
        if os.path.isdir(git_mirror_repodir):
            common.local_rsync(options, [git_mirror_repodir + '/'], 'repo/')
        if os.path.isdir(git_mirror_metadatadir):
            common.local_rsync(options, [git_mirror_metadatadir + '/'], 'metadata/')
        if os.path.isdir(git_mirror_statsdir):
            common.local_rsync(options, [git_mirror_statsdir + '/'], 'stats/')

        ssh_private_key_file = _ssh_key_from_debug_keystore()
        # this is needed for GitPython to find the SSH key
        ssh_dir = os.path.join(os.getenv('HOME'), '.ssh')
        os.makedirs(ssh_dir, exist_ok=True)
        ssh_config = os.path.join(ssh_dir, 'config')
        logging.debug(_('adding IdentityFile to {path}').format(path=ssh_config))
        with open(ssh_config, 'a') as fp:
            fp.write('\n\nHost *\n\tIdentityFile %s\n' % ssh_private_key_file)

        if options.archive_older == archive_older_unset:
            fdroid_size = common.get_dir_size(git_mirror_fdroiddir)
            max_size = common.GITLAB_COM_PAGES_MAX_SIZE
            if fdroid_size < max_size:
                options.archive_older = 20
            else:
                options.archive_older = 3
                print(
                    'WARNING: repo is %s over the GitLab Pages limit (%s)'
                    % (fdroid_size - max_size, max_size)
                )
                print('Setting --archive-older to 3')

        config = {
            'identity_file': ssh_private_key_file,
            'repo_name': repo_git_base,
            'repo_url': repo_url,
            'repo_description': 'Nightly builds from %s' % git_user_email,
            'archive_name': repo_git_base + ' archive',
            'archive_url': repo_base + '/archive',
            'archive_description': 'Old nightly builds that have been archived.',
            'archive_older': options.archive_older,
            'servergitmirrors': [{"url": servergitmirror}],
            'keystore': KEYSTORE_FILE,
            'repo_keyalias': KEY_ALIAS,
            'keystorepass': PASSWORD,
            'keypass': PASSWORD,
            'keydname': DISTINGUISHED_NAME,
            'make_current_version_link': False,
        }
        with open('config.yml', 'w') as fp:
            yaml.dump(config, fp, default_flow_style=False)
        os.chmod('config.yml', 0o600)
        config = common.read_config()
        common.assert_config_keystore(config)

        logging.debug(
            _('Run over {cibase} to find -debug.apk. and skip repo_basedir {repo_basedir}').format(
                cibase=cibase,
                repo_basedir=repo_basedir
            )
        )

        for root, dirs, files in os.walk(cibase):
            for d in ('.git', '.gradle'):
                if d in dirs:
                    dirs.remove(d)
            if root == cibase and 'fdroid' in dirs:
                dirs.remove('fdroid')

            for f in files:
                if f.endswith('-debug.apk'):
                    apkfilename = os.path.join(root, f)
                    logging.debug(
                        _('Stripping mystery signature from {apkfilename}').format(
                            apkfilename=apkfilename
                        )
                    )
                    destapk = os.path.join(repodir, os.path.basename(f))
                    os.chmod(apkfilename, 0o644)
                    logging.debug(
                        _(
                            'Resigning {apkfilename} with provided debug.keystore'
                        ).format(apkfilename=os.path.basename(apkfilename))
                    )
                    common.sign_apk(apkfilename, destapk, KEY_ALIAS)

        if options.verbose:
            logging.debug(_('attempting bare SSH connection to test deploy key:'))
            try:
                subprocess.check_call(
                    [
                        'ssh',
                        '-Tvi',
                        ssh_private_key_file,
                        '-oIdentitiesOnly=yes',
                        '-oStrictHostKeyChecking=no',
                        servergitmirror.split(':')[0],
                    ]
                )
            except subprocess.CalledProcessError:
                pass

        app_url = clone_url[: -len(NIGHTLY)]
        template = dict()
        template['AuthorName'] = clone_url.split('/')[4]
        template['AuthorWebSite'] = '/'.join(clone_url.split('/')[:4])
        template['Categories'] = ['nightly']
        template['SourceCode'] = app_url
        template['IssueTracker'] = app_url + '/issues'
        template['Summary'] = 'Nightly build of ' + urlparse(app_url).path[1:]
        template['Description'] = template['Summary']
        with open('template.yml', 'w') as fp:
            yaml.dump(template, fp)

        subprocess.check_call(
            ['fdroid', 'update', '--rename-apks', '--create-metadata', '--verbose'],
            cwd=repo_basedir,
        )
        common.local_rsync(
            options, [repo_basedir + '/metadata/'], git_mirror_metadatadir + '/'
        )
        stats = repo_basedir + '/stats/'
        if os.path.exists(stats):
            common.local_rsync(options, [stats], git_mirror_statsdir + '/')
        mirror_git_repo.git.add(all=True)
        mirror_git_repo.index.commit("update app metadata")

        if not options.no_deploy:
            try:
                cmd = ['fdroid', 'deploy', '--verbose', '--no-keep-git-mirror-archive']
                subprocess.check_call(cmd, cwd=repo_basedir)
            except subprocess.CalledProcessError:
                logging.error(
                    _('cannot publish update, did you set the deploy key?')
                    + '\n'
                    + deploy_key_url
                )
                sys.exit(1)

        if not options.keep_private_keys:
            os.remove(KEYSTORE_FILE)
            if shutil.rmtree.avoids_symlink_attacks:
                shutil.rmtree(os.path.dirname(ssh_private_key_file))

    else:
        if not os.path.isfile(options.keystore):
            androiddir = os.path.dirname(options.keystore)
            if not os.path.exists(androiddir):
                os.mkdir(androiddir)
                logging.info(_('created {path}').format(path=androiddir))
            logging.error(_('{path} does not exist!  Create it by running:').format(path=options.keystore)
                          + '\n    keytool -genkey -v -keystore ' + options.keystore + ' -storepass android \\'
                          + '\n     -alias androiddebugkey -keypass android -keyalg RSA -keysize 2048 -validity 10000 \\'
                          + '\n     -dname "CN=Android Debug,O=Android,C=US"')
            sys.exit(1)
        ssh_dir = os.path.join(os.getenv('HOME'), '.ssh')
        privkey = _ssh_key_from_debug_keystore(options.keystore)
        if os.path.exists(ssh_dir):
            ssh_private_key_file = os.path.join(ssh_dir, os.path.basename(privkey))
            shutil.move(privkey, ssh_private_key_file)
            shutil.move(privkey + '.pub', ssh_private_key_file + '.pub')
        if shutil.rmtree.avoids_symlink_attacks:
            shutil.rmtree(os.path.dirname(privkey))

        if options.show_secret_var:
            debug_keystore = _get_keystore_secret_var(options.keystore)
            print(
                _('\n{path} encoded for the DEBUG_KEYSTORE secret variable:').format(
                    path=options.keystore
                )
            )
            print(debug_keystore)

    os.umask(umask)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os
import unittest

from git import Repo

import fdroidserver.common
import fdroidserver.metadata
from .testcommon import mkdtemp, VerboseFalseOptions


class VCSTest(unittest.TestCase):
    """For some reason the VCS classes are in fdroidserver/common.py"""

    def setUp(self):
        self._td = mkdtemp()
        os.chdir(self._td.name)

    def tearDown(self):
        self._td.cleanup()

    def test_remote_set_head_can_fail(self):
        # First create an upstream repo with one commit
        upstream_repo = Repo.init("upstream_repo")
        with open(upstream_repo.working_dir + "/file", 'w') as f:
            f.write("Hello World!")
        upstream_repo.index.add([upstream_repo.working_dir + "/file"])
        upstream_repo.index.commit("initial commit")
        commitid = upstream_repo.head.commit.hexsha

        # Now clone it once manually, like gitlab runner gitlab-runner sets up a repo during CI
        clone1 = Repo.init("clone1")
        clone1.create_remote("upstream", "file://" + upstream_repo.working_dir)
        clone1.remote("upstream").fetch()
        clone1.head.reference = clone1.commit(commitid)
        clone1.head.reset(index=True, working_tree=True)
        self.assertTrue(clone1.head.is_detached)

        # and now we want to use this clone as a source repo for fdroid build
        config = {}
        os.mkdir("build")
        config['sdk_path'] = 'MOCKPATH'
        config['ndk_paths'] = {'r10d': os.getenv('ANDROID_NDK_HOME')}
        config['java_paths'] = {'fake': 'fake'}
        fdroidserver.common.config = config
        app = fdroidserver.metadata.App()
        app.RepoType = 'git'
        app.Repo = clone1.working_dir
        app.id = 'com.gpl.rpg.AndorsTrail'
        build = fdroidserver.metadata.Build()
        build.commit = commitid
        build.androidupdate = ['no']
        vcs, build_dir = fdroidserver.common.setup_vcs(app)
        # force an init of the repo, the remote head error only occurs on the second gotorevision call

        fdroidserver.common.options = VerboseFalseOptions
        vcs.gotorevision(build.commit)
        fdroidserver.common.prepare_source(
            vcs,
            app,
            build,
            build_dir=build_dir,
            srclib_dir="ignore",
            extlib_dir="ignore",
        )
        self.assertTrue(os.path.isfile("build/com.gpl.rpg.AndorsTrail/file"))



#!/usr/bin/env python3

import unittest
import unittest.mock

from .testcommon import mock_urlopen
import fdroidserver


class GithubApiTest(unittest.TestCase):
    def test__init(self):
        api = fdroidserver.github.GithubApi('faketoken', 'fakerepopath')
        self.assertEqual(api._api_token, 'faketoken')
        self.assertEqual(api._repo_path, 'fakerepopath')

    def test__req(self):
        api = fdroidserver.github.GithubApi('faketoken', 'fakerepopath')
        r = api._req('https://fakeurl', data='fakedata')
        self.assertEqual(r.full_url, 'https://fakeurl')
        self.assertEqual(r.data, "fakedata")
        self.assertDictEqual(
            r.headers,
            {
                'Accept': 'application/vnd.github+json',
                'Authorization': 'Bearer faketoken',
                'X-github-api-version': '2022-11-28',
            },
        )

    def test_list_released_tags(self):
        api = fdroidserver.github.GithubApi('faketoken', 'fakerepopath')
        uomock = mock_urlopen(
            body='[{"tag_name": "fake"}, {"tag_name": "double_fake"}]'
        )
        with unittest.mock.patch("urllib.request.urlopen", uomock):
            result = api.list_released_tags()
        self.assertListEqual(result, ['fake', 'double_fake'])

    def test_list_unreleased_tags(self):
        api = fdroidserver.github.GithubApi('faketoken', 'fakerepopath')

        api.list_all_tags = unittest.mock.Mock(return_value=[1, 2, 3, 4])
        api.list_released_tags = unittest.mock.Mock(return_value=[1, 2])

        result = api.list_unreleased_tags()

        self.assertListEqual(result, [3, 4])

    def test_tag_exists(self):
        api = fdroidserver.github.GithubApi('faketoken', 'fakerepopath')
        uomock = mock_urlopen(body='[{"ref": "refs/tags/fake_tag"}]')
        with unittest.mock.patch("urllib.request.urlopen", uomock):
            result = api.tag_exists('fake_tag')
        self.assertTrue(result)

    def test_tag_exists_failure(self):
        api = fdroidserver.github.GithubApi('faketoken', 'fakerepopath')

        uomock = mock_urlopen(body='[{"error": "failure"}]')

        with unittest.mock.patch("urllib.request.urlopen", uomock):
            success = api.tag_exists('fake_tag')

        self.assertFalse(success)

    def test_list_all_tags(self):
        api = fdroidserver.github.GithubApi('faketoken', 'fakerepopath')

        uomock = mock_urlopen(
            body='[{"ref": "refs/tags/fake"}, {"ref": "refs/tags/double_fake"}]'
        )

        with unittest.mock.patch("urllib.request.urlopen", uomock):
            result = api.list_all_tags()

        self.assertListEqual(result, ['fake', 'double_fake'])

    def test_create_release(self):
        api = fdroidserver.github.GithubApi('faketoken', 'fakerepopath')

        uomock = mock_urlopen(body='{"id": "fakeid"}')
        api.tag_exists = lambda x: True
        api._create_release_asset = unittest.mock.Mock()

        with unittest.mock.patch("urllib.request.urlopen", uomock):
            success = api.create_release('faketag', ['file_a', 'file_b'], body="bdy")
        self.assertTrue(success)

        req = uomock.call_args_list[0][0][0]
        self.assertEqual(1, len(uomock.call_args_list))
        self.assertEqual(2, len(uomock.call_args_list[0]))
        self.assertEqual(1, len(uomock.call_args_list[0][0]))
        self.assertEqual(
            req.full_url,
            'https://api.github.com/repos/fakerepopath/releases',
        )
        self.assertEqual(req.data, b'{"tag_name": "faketag", "body": "bdy"}')
        self.assertListEqual(
            api._create_release_asset.call_args_list,
            [
                unittest.mock.call('fakeid', 'file_a'),
                unittest.mock.call('fakeid', 'file_b'),
            ],
        )

    def test__create_release_asset(self):
        api = fdroidserver.github.GithubApi('faketoken', 'fakerepopath')
        uomock = mock_urlopen()

        with unittest.mock.patch(
            'fdroidserver.github.open',
            unittest.mock.mock_open(read_data=b"fake_content"),
        ), unittest.mock.patch("urllib.request.urlopen", uomock):
            success = api._create_release_asset('fake_id', 'fake_file')

        self.assertTrue(success)

        req = uomock.call_args_list[0][0][0]
        self.assertEqual(1, len(uomock.call_args_list))
        self.assertEqual(2, len(uomock.call_args_list[0]))
        self.assertEqual(1, len(uomock.call_args_list[0][0]))
        self.assertEqual(
            req.full_url,
            'https://uploads.github.com/repos/fakerepopath/releases/fake_id/assets?name=fake_file',
        )
        self.assertDictEqual(
            req.headers,
            {
                "Accept": "application/vnd.github+json",
                'Authorization': 'Bearer faketoken',
                'Content-type': 'application/octet-stream',
                'X-github-api-version': '2022-11-28',
            },
        )
        self.assertEqual(req.data, b'fake_content')



#!/usr/bin/env python3
#
# an fdroid plugin for setting up srclibs
#
# The 'fdroid build' gitlab-ci job uses --on-server, which does not
# set up the srclibs.  This plugin does the missing setup.

import glob
import os
from argparse import ArgumentParser
from fdroidserver import _, common, scanner


fdroid_summary = 'Run scanner.scan_binary on APKs'


def main():
    parser = ArgumentParser()
    common.setup_global_opts(parser)
    parser.add_argument("APK", nargs='*', help=_("Path to a signed or unsigned APK."))
    options = common.parse_args(parser)
    common.read_config()
    if options.APK:
        apks = options.APK
    else:
        apks = glob.glob(os.path.join('unsigned', '*.apk'))

    errors = 0
    for apk in apks:
        print('Scanning', apk, '...')
        if scanner.scan_binary(apk):
            print("ERROR: Found blocklisted packages in:", apk)
            errors += 1
    exit(errors)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# net.py - part of the FDroid server tools
# Copyright (C) 2015 Hans-Christoph Steiner <hans@eds.org>
# Copyright (C) 2022 FC Stegerman <flx@obfusk.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import logging
import os
import random
import requests
import tempfile
import time
import urllib
import urllib3
from requests.adapters import HTTPAdapter, Retry

from . import _, common

logger = logging.getLogger(__name__)

HEADERS = {'User-Agent': 'F-Droid'}


def download_file(url, local_filename=None, dldir='tmp', retries=3, backoff_factor=0.1):
    """Try hard to download the file, including retrying on failures.

    This has two retry cycles, one inside of the requests session, the
    other provided by this function.  The requests retry logic applies
    to failed DNS lookups, socket connections and connection timeouts,
    never to requests where data has made it to the server. This
    handles ChunkedEncodingError during transfer in its own retry
    loop.  This can result in more retries than are specified in the
    retries parameter.

    """
    filename = urllib.parse.urlparse(url).path.split('/')[-1]
    if local_filename is None:
        local_filename = os.path.join(dldir, filename)
    for i in range(retries + 1):
        if retries:
            max_retries = Retry(total=retries - i, backoff_factor=backoff_factor)
            adapter = HTTPAdapter(max_retries=max_retries)
            session = requests.Session()
            session.mount('http://', adapter)
            session.mount('https://', adapter)
        else:
            session = requests
        # the stream=True parameter keeps memory usage low
        r = session.get(
            url, stream=True, allow_redirects=True, headers=HEADERS, timeout=300
        )
        r.raise_for_status()
        try:
            with open(local_filename, 'wb') as f:
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
                        f.flush()
            return local_filename
        except requests.exceptions.ChunkedEncodingError as err:
            if i == retries:
                raise err
            logger.warning('Download interrupted, retrying...')
            time.sleep(backoff_factor * 2**i)
    raise ValueError("retries must be >= 0")


def download_using_mirrors(mirrors, local_filename=None):
    """Try to download the file from any working mirror.

    Download the file that all URLs in the mirrors list point to,
    trying all the tricks, starting with the most private methods
    first.  The list of mirrors is converted into a list of mirror
    configurations to try, in order that the should be attempted.

    This builds mirror_configs_to_try using all possible combos to
    try.  If a mirror is marked with worksWithoutSNI: True, then this
    logic will try it twice: first without SNI, then again with SNI.

    """
    mirrors = common.parse_mirrors_config(mirrors)
    mirror_configs_to_try = []
    for mirror in mirrors:
        mirror_configs_to_try.append(mirror)
        if mirror.get('worksWithoutSNI'):
            m = copy.deepcopy(mirror)
            del m['worksWithoutSNI']
            mirror_configs_to_try.append(m)

    if not local_filename:
        for mirror in mirrors:
            filename = urllib.parse.urlparse(mirror['url']).path.split('/')[-1]
            if filename:
                break
        if filename:
            local_filename = os.path.join(common.get_cachedir(), filename)
        else:
            local_filename = tempfile.mkstemp(prefix='fdroid-')

    timeouts = (2, 10, 100)
    last_exception = None
    for timeout in timeouts:
        for mirror in mirror_configs_to_try:
            last_exception = None
            urllib3.util.ssl_.HAS_SNI = not mirror.get('worksWithoutSNI')
            try:
                # the stream=True parameter keeps memory usage low
                r = requests.get(
                    mirror['url'],
                    stream=True,
                    allow_redirects=False,
                    headers=HEADERS,
                    # add jitter to the timeout to be less predictable
                    timeout=timeout + random.randint(0, timeout),  # nosec B311
                )
                if r.status_code != 200:
                    raise requests.exceptions.HTTPError(r.status_code, response=r)
                with open(local_filename, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=1024):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)
                            f.flush()
                return local_filename
            except (
                ConnectionError,
                requests.exceptions.ChunkedEncodingError,
                requests.exceptions.ConnectionError,
                requests.exceptions.ContentDecodingError,
                requests.exceptions.HTTPError,
                requests.exceptions.SSLError,
                requests.exceptions.StreamConsumedError,
                requests.exceptions.Timeout,
                requests.exceptions.UnrewindableBodyError,
            ) as e:
                last_exception = e
                logger.debug(_('Retrying failed download: %s') % str(e))
    # if it hasn't succeeded by now, then give up and raise last exception
    if last_exception:
        raise last_exception


def http_get(url, etag=None, timeout=600):
    """Download the content from the given URL by making a GET request.

    If an ETag is given, it will do a HEAD request first, to see if the content changed.

    Parameters
    ----------
    url
      The URL to download from.
    etag
      The last ETag to be used for the request (optional).

    Returns
    -------
    A tuple consisting of:
      - The raw content that was downloaded or None if it did not change
      - The new eTag as returned by the HTTP request
    """
    # TODO disable TLS Session IDs and TLS Session Tickets
    #      (plain text cookie visible to anyone who can see the network traffic)
    if etag:
        r = requests.head(url, headers=HEADERS, timeout=timeout)
        r.raise_for_status()
        if 'ETag' in r.headers and etag == r.headers['ETag']:
            return None, etag

    r = requests.get(url, headers=HEADERS, timeout=timeout)
    r.raise_for_status()

    new_etag = None
    if 'ETag' in r.headers:
        new_etag = r.headers['ETag']

    return r.content, new_etag



#!/usr/bin/env python3

import os
import textwrap
import unittest

from pathlib import Path
from unittest.mock import Mock, patch

import fdroidserver
from fdroidserver import common, install
from fdroidserver.exception import BuildException, FDroidException


class InstallTest(unittest.TestCase):
    '''fdroidserver/install.py'''

    def tearDown(self):
        common.config = None

    def test_devices(self):
        config = dict()
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.common.config = config
        try:
            config['adb'] = fdroidserver.common.find_sdk_tools_cmd('adb')
        except FDroidException as e:
            self.skipTest(f'Skipping test because: {e}')
        self.assertTrue(os.path.exists(config['adb']))
        self.assertTrue(os.path.isfile(config['adb']))
        devices = fdroidserver.install.devices()
        self.assertIsInstance(devices, list, 'install.devices() did not return a list!')
        for device in devices:
            self.assertIsInstance(device, str)

    def test_devices_fail(self):
        common.config = dict()
        common.fill_config_defaults(common.config)
        common.config['adb'] = '/bin/false'
        with self.assertRaises(FDroidException):
            fdroidserver.install.devices()

    def test_devices_fail_nonexistent(self):
        """This is mostly just to document this strange difference in behavior"""
        common.config = dict()
        common.fill_config_defaults(common.config)
        common.config['adb'] = '/nonexistent'
        with self.assertRaises(BuildException):
            fdroidserver.install.devices()

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_none(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = 'List of devices attached\n\n'
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual([], fdroidserver.install.devices())

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_one(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = 'List of devices attached\n05995813\tdevice\n\n'
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual(['05995813'], fdroidserver.install.devices())

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_many(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = textwrap.dedent(
            """* daemon not running; starting now at tcp:5037
            * daemon started successfully
            List of devices attached
            RZCT809FTQM	device
            05995813	device
            emulator-5556	device
            emulator-5554	unauthorized
            0a388e93	no permissions (missing udev rules? user is in the plugdev group); see [http://developer.android.com/tools/device.html]
            986AY133QL	device
            09301JEC215064	device
            015d165c3010200e	device
            4DCESKVGUC85VOTO	device

            """
        )
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual(
            [
                'RZCT809FTQM',
                '05995813',
                'emulator-5556',
                '986AY133QL',
                '09301JEC215064',
                '015d165c3010200e',
                '4DCESKVGUC85VOTO',
            ],
            fdroidserver.install.devices(),
        )

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_error(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = textwrap.dedent(
            """* daemon not running. starting it now on port 5037 *
            * daemon started successfully *
            ** daemon still not running
            error: cannot connect to daemon
            """
        )
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual([], fdroidserver.install.devices())

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_no_permissions(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = textwrap.dedent(
            """List of devices attached
            ????????????????	no permissions
            """
        )
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual([], fdroidserver.install.devices())

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_unauthorized(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = textwrap.dedent(
            """List of devices attached
            aeef5e4e	unauthorized
            """
        )
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual([], fdroidserver.install.devices())

    @patch('fdroidserver.common.SdkToolsPopen')
    def test_devices_with_mock_no_permissions_with_serial(self, mock_SdkToolsPopen):
        p = Mock()
        mock_SdkToolsPopen.return_value = p
        p.output = textwrap.dedent(
            """List of devices attached
             4DCESKVGUC85VOTO	no permissions (missing udev rules? user is in the plugdev group); see [http://developer.android.com/tools/device.html]

            """
        )
        p.returncode = 0
        common.config = dict()
        common.fill_config_defaults(common.config)
        self.assertEqual([], fdroidserver.install.devices())

    @staticmethod
    def _download_raise(privacy_mode):
        raise Exception('fake failed download')

    @patch('fdroidserver.install.download_apk')
    @patch('fdroidserver.install.download_fdroid_apk')
    @patch('fdroidserver.install.download_fdroid_apk_from_github')
    @patch('fdroidserver.install.download_fdroid_apk_from_ipns')
    @patch('fdroidserver.install.download_fdroid_apk_from_maven')
    def test_install_fdroid_apk_privacy_mode_true(
        self, maven, ipns, github, download_fdroid_apk, download_apk
    ):
        download_apk.side_effect = self._download_raise
        download_fdroid_apk.side_effect = self._download_raise
        github.side_effect = self._download_raise
        ipns.side_effect = self._download_raise
        maven.side_effect = self._download_raise
        fdroidserver.common.config = {'jarsigner': 'fakepath'}
        install.install_fdroid_apk(privacy_mode=True)
        download_apk.assert_not_called()
        download_fdroid_apk.assert_not_called()
        github.assert_called_once()
        ipns.assert_called_once()
        maven.assert_called_once()

    @patch('fdroidserver.install.download_apk')
    @patch('fdroidserver.install.download_fdroid_apk')
    @patch('fdroidserver.install.download_fdroid_apk_from_github')
    @patch('fdroidserver.install.download_fdroid_apk_from_ipns')
    @patch('fdroidserver.install.download_fdroid_apk_from_maven')
    def test_install_fdroid_apk_privacy_mode_false(
        self, maven, ipns, github, download_fdroid_apk, download_apk
    ):
        download_apk.side_effect = self._download_raise
        download_fdroid_apk.side_effect = self._download_raise
        github.side_effect = self._download_raise
        ipns.side_effect = self._download_raise
        maven.side_effect = self._download_raise
        fdroidserver.common.config = {'jarsigner': 'fakepath'}
        install.install_fdroid_apk(privacy_mode=False)
        download_apk.assert_called_once()
        download_fdroid_apk.assert_called_once()
        github.assert_called_once()
        ipns.assert_called_once()
        maven.assert_called_once()

    @patch('fdroidserver.install.download_apk')
    @patch('fdroidserver.install.download_fdroid_apk')
    @patch('fdroidserver.install.download_fdroid_apk_from_github')
    @patch('fdroidserver.install.download_fdroid_apk_from_ipns')
    @patch('fdroidserver.install.download_fdroid_apk_from_maven')
    @patch('locale.getlocale', lambda: ('zh_CN', 'UTF-8'))
    def test_install_fdroid_apk_privacy_mode_locale_auto(
        self, maven, ipns, github, download_fdroid_apk, download_apk
    ):
        download_apk.side_effect = self._download_raise
        download_fdroid_apk.side_effect = self._download_raise
        github.side_effect = self._download_raise
        ipns.side_effect = self._download_raise
        maven.side_effect = self._download_raise
        fdroidserver.common.config = {'jarsigner': 'fakepath'}
        install.install_fdroid_apk(privacy_mode=None)
        download_apk.assert_not_called()
        download_fdroid_apk.assert_not_called()
        github.assert_called_once()
        ipns.assert_called_once()
        maven.assert_called_once()

    @patch('fdroidserver.net.download_using_mirrors', lambda m: 'testvalue')
    def test_download_fdroid_apk_smokecheck(self):
        self.assertEqual('testvalue', install.download_fdroid_apk())

    @unittest.skipUnless(os.getenv('test_download_fdroid_apk'), 'requires net access')
    def test_download_fdroid_apk(self):
        f = install.download_fdroid_apk()
        self.assertTrue(Path(f).exists())

    @unittest.skipUnless(os.getenv('test_download_fdroid_apk'), 'requires net access')
    def test_download_fdroid_apk_from_maven(self):
        f = install.download_fdroid_apk_from_maven()
        self.assertTrue(Path(f).exists())

    @unittest.skipUnless(os.getenv('test_download_fdroid_apk'), 'requires net access')
    def test_download_fdroid_apk_from_ipns(self):
        f = install.download_fdroid_apk_from_ipns()
        self.assertTrue(Path(f).exists())

    @unittest.skipUnless(os.getenv('test_download_fdroid_apk'), 'requires net access')
    def test_download_fdroid_apk_from_github(self):
        f = install.download_fdroid_apk_from_github()
        self.assertTrue(Path(f).exists())



#!/usr/bin/env python3

import ruamel.yaml

from pathlib import Path

mirrors_yml = Path('/home/hans/code/fdroid/fdroiddata/config/mirrors.yml')
with mirrors_yml.open() as fp:
    mirrors_config = ruamel.yaml.YAML(typ='safe').load(fp)

for d in mirrors_config:
    d['url'] += '/repo'
    print(d, end=',\n')
